/* ***** BEGIN LICENSE BLOCK *****
 *
 * Copyright (c) 2011 Groupe LINAGORA
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of the
 *  License, (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  http://www.obm.org/
 *
 * ***** END LICENSE BLOCK ***** */

/*
 This file defines all asynchronous task to perform synchronization
 of addressbook data between lightning and obm-sync.
 These tasks should be used with the Scheduler object.
*/


/**
 * asynchronous task to retrieve changes from the server
 */
function getAddressbookSyncTask(abSync, isTheFirstPass) {
	this.isTheFirstPass = isTheFirstPass;
	this.taskName = "getAddressbookSyncTask";
	this.wsMethod = 'book/getAddressBookSync';
	this.abSync = abSync;
}

getAddressbookSyncTask.prototype = new xhrTask();

getAddressbookSyncTask.prototype.preRequest = function(data, parameters) {
	this.abSync.callerListener.onNewOperation("getAddressBookSync", null);
	this.abSync.callerListener.onProgress(0, 1, "", "contacts");
	//this.logger.debug("run task "+(this.isTheFirstPass?"first pass":"second pass"), false, false);
	parameters.add("sid", this.abSync.session_id);
	parameters.add("lastSync", this.abSync._getLastSync());
};

getAddressbookSyncTask.prototype.postRequest = function(callback, data, xml, error) {
	if (error) {
		callback(this._processError(error));
		return;
	}
	var ab = new AddressBookChangesResponseShell(xml);
	this.abSync._onGetAddressBookSync(this.isTheFirstPass, ab);

	this.abSync.callerListener.onOperationDone(true, null);

	this.abSync._mergeRemoteDiff();
	callback();
};


/**
 * synchronous task which prepare various asynchronous task
 * to send local modification to the server
 */
function abSendLocalDiffTask(abSync) {
	this.taskName = "abSendLocalDiffTask";
	this.abSync = abSync;
}

abSendLocalDiffTask.prototype = new basicTask();

abSendLocalDiffTask.prototype.run = function(callback, data) {

    var queriesCount = this.abSync.getLocalDiffCount();

    if (queriesCount == 0) {
        callback();
    }
    else {
        this.abSync.callerListener.onNewOperation("sendLocalDiff", queriesCount);

		var tasks = [];
        var listLength;
		for (var i = 0; i < this.abSync.abNames.length; i++) {
			var abName = this.abSync.abNames[i];
			listLength = this.abSync._localDiff[abName].removed.length;
			for (var j = 0; j < listLength; j++) {
				var task = new removeContactTask(this.abSync,
												 abName,
												 this.abSync._localDiff[abName].removed[j],
												 j,
												 listLength);
				tasks.push(task);
			}
			listLength = this.abSync._localDiff[abName].added.length;
			for (var j = 0; j < listLength; j++) {
				var task = new addContactTask(this.abSync,
											  abName,
											  this.abSync._localDiff[abName].added[j],
											  j,
											  listLength);
				tasks.push(task);
			}
			listLength = this.abSync._localDiff[abName].updated.length;
			for (var j = 0; j < listLength; j++) {
				var task = new updateContactTask(this.abSync,
												 abName,
												 this.abSync._localDiff[abName].updated[j],
												 j,
												 listLength);
				tasks.push(task);
			}
		}

		listLength = this.abSync._localDiff.unsubscribed.length;
		for (var j = 0; j < listLength; j++) {
			var task = new unsubscribeBookTask(this.abSync,
											   this.abSync._localDiff.unsubscribed[j],
											   j,
											   listLength);
			tasks.push(task);
		}

        var self = this;
        // an additionnal task to close the operation.
        var sendLocalDiffEndTask = {
			taskName: "abSendLocalDiffEndTask",
            run: function(callback, data) {
                self.abSync.callerListener.onOperationDone(true, null);
                callback();
            },
            abort: function() {}
        }

		tasks.push(sendLocalDiffEndTask);

		callback(null, {tasks:tasks});
    }
};

/**
 * simple task to send a contact removing to the server
 */
function removeContactTask(abSync, abName, contactId, idx, total) {
	this.abName = abName;
	this.idx = idx;
	this.total = total;
	this.contactId = contactId;
	this.taskName = "removeContactTask";
	this.wsMethod = 'book/removeContact';
	this.abSync = abSync;
}

removeContactTask.prototype = new xhrTask();

removeContactTask.prototype.preRequest = function(data, parameters) {
	this.abSync.callerListener.onProgress(this.idx, this.total, "removed", this.abName);
	parameters.add("sid",  this.abSync.session_id);
	parameters.add("book", (this.abName == "users"? this.abName: "contacts"));
	parameters.add("id",   this.contactId);
};

removeContactTask.prototype.postRequest = function(callback, data, xml, error) {
	if (error) {
		this.abSync.callerListener.onError(this._processError(error));
	}
	else {
		UtilsOBM.removeItemFromChangesList(Ci.nsIAbDirectory, this.abName, Ci.nsIAbCard, "removed", this.contactId);
	}
	callback();
};


/**
 * simple task to send a contact creation to the server
 */
function addContactTask(abSync, abName, contactId, idx, total) {
    this.abName = abName;
    this.idx = idx;
    this.total = total;
    this.contactId = contactId;
	this.taskName = "addContactTask";
	this.wsMethod = 'book/createContactInBook';
	this.abSync = abSync;
}

addContactTask.prototype = new xhrTask();

addContactTask.prototype.preRequest = function(data, p) {
	this.abSync.callerListener.onProgress(this.idx, this.total, "added", this.abName);

	var contactToSend = this.abSync._createContactObjFromId(this.abName, this.contactId, true);
	p.add("sid",  this.abSync.session_id);
	p.add("bookId", this.abName);
	p.add("contact", JSObjContactShell(contactToSend));
};

addContactTask.prototype.postRequest = function(callback, data, xml, error) {
	if (error) {
		this.abSync.callerListener.onError(this._processError(error));
	}
	else {
		var contact = new ContactShell(xml);

		var updatedContact = contact.QueryInterface(Ci.obmIRContact);

		// TODO. OBM_NEW_AD_MANAGER. use obmIAddressBooksManager to get a local contact
		var localContact = this.abSync._getLocalContact(this.abName, this.contactId);
		this.abSync._updateLocalContact(localContact, updatedContact);
		// TODO. OBM_NEW_AD_MANAGER. use obmIAddressBooksManager to update a local contact
		if (this.abSync.abVersion != 3) {
			var addrDb = this.abSync._getDatabaseFromABName(this.abName);
			addrDb.editCard(localContact, false, null);
			addrDb.commit(Ci.nsAddrDBCommitType.kSessionCommit);
		} else {
			var directory = this.abSync._getDirectoryFromABName(this.abName);
			directory.modifyCard(localContact);
		}

		UtilsOBM.removeItemFromChangesList(Ci.nsIAbDirectory, this.abName, Ci.nsIAbCard, "added", this.contactId);
	}
	callback();
};


/**
 * simple task to send a contact change to the server
 */
function updateContactTask(abSync, abName, contactId, idx, total) {
    this.abName = abName;
    this.idx = idx;
    this.total = total;
    this.contactId = contactId;
	this.taskName = "updateContactTask";
	this.wsMethod = 'book/modifyContact';
	this.abSync = abSync;
}

updateContactTask.prototype = new xhrTask();

updateContactTask.prototype.preRequest = function(data, parameters) {

    this.abSync.callerListener.onProgress(this.idx, this.total, "updated", this.abName);

    var contactToSend = this.abSync._createContactObjFromId(this.abName, this.contactId, false);

	parameters.add("sid",  this.abSync.session_id);
    parameters.add("book", (this.abName == "users"? this.abName: "contacts"));
    parameters.add("contact", JSObjContactShell(contactToSend));
};

updateContactTask.prototype.postRequest = function(callback, data, xml, error) {
	if (error) {
		this.abSync.callerListener.onError(this._processError(error));
	}
	else {
		var contact = new ContactShell(xml);

		var updatedContact = contact.QueryInterface(Ci.obmIRContact);

		// TODO. OBM_NEW_AD_MANAGER. use obmIAddressBooksManager to get a local contact
		var localContact = this.abSync._getLocalContact(this.abName, this.contactId);
		this.abSync._updateLocalContact(localContact, updatedContact);
		// TODO. OBM_NEW_AD_MANAGER. use obmIAddressBooksManager to update a local contact
		if (this.abSync.abVersion != 3) {
			var addrDb = this.abSync._getDatabaseFromABName(this.abName);
			addrDb.editCard(localContact, false, null);
			addrDb.commit(Ci.nsAddrDBCommitType.kSessionCommit);
		} else {
			var directory = this.abSync._getDirectoryFromABName(this.abName);
			directory.modifyCard(localContact);
		}

		UtilsOBM.removeItemFromChangesList(Ci.nsIAbDirectory, this.abName, Ci.nsIAbCard, "updated", this.contactId);
	}
	callback();
};


/**
 * simple task to send a book unsubscribtion to the server
 */
function unsubscribeBookTask(abSync, abName, idx, total) {
    this.abName = abName;
    this.idx = idx;
    this.total = total;
	this.taskName = "unsubscribeBookTask";
	this.wsMethod = 'book/unsubscribeBook';
	this.abSync = abSync;
}

unsubscribeBookTask.prototype = new xhrTask();

unsubscribeBookTask.prototype.preRequest = function(data, p) {
    this.abSync.callerListener.onProgress(this.idx, this.total, "unsubscribed", this.abName);
	p.add("sid",  this.abSync.session_id);
    p.add("bookId", this.abName);
};

unsubscribeBookTask.prototype.postRequest = function(callback, data, xml, error) {

	if (error) {
		this.abSync.callerListener.onError(this._processError(error));
	}
	else {
		// TODO: handle boolean return
		UtilsOBM.removeItemFromChangesList(Ci.nsIAbDirectory, "list", Ci.nsIAbDirectory, "removed", this.abName);
	}

	callback();
};


// --------------- utilities objects to read response data
//
var _obmTypesMappings = {};
_obmTypesMappings["CELL;VOICE;X-OBM-Ref1"] = "mobilePhone";
_obmTypesMappings["PAGER;X-OBM-Ref1"] = "pagerPhone";
_obmTypesMappings["WORK;VOICE;X-OBM-Ref1"] = "workPhone";
_obmTypesMappings["WORK;FAX;X-OBM-Ref1"] = "workFax";
_obmTypesMappings["HOME;VOICE;X-OBM-Ref1"] = "homePhone";
_obmTypesMappings["INTERNET;X-OBM-Ref1"] = "email";
_obmTypesMappings["INTERNET;X-OBM-Ref2"] = "email2";
_obmTypesMappings["URL;X-OBM-Ref1"] = "workWebPage";
_obmTypesMappings["URL;X-OBM-Ref2"] = "homeWebPage";
_obmTypesMappings["HOME;X-OBM-Ref1"] = "home";
_obmTypesMappings["WORK;X-OBM-Ref1"] = "work";

function AddressBookShell(aXml) {
	this._mXml = aXml;
}

AddressBookShell.prototype = {
	QueryInterface: function(aIID) {
		if (!aIID.equals(Ci.nsISupports) &&
			!aIID.equals(Ci.obmIRAddressBook)) {
			throw Cr.NS_ERROR_NO_INTERFACE;
		}
		return this;
	},
	get uid() {return _xL(this._mXml.@uid)},
	get	name() {return _xS(this._mXml.*::name)},
	get	ownerDisplayName() {return _xS(this._mXml.*::ownerDisplayName)}
}

function ContactShell(aXml) {
	this._mXml = aXml;
	
	this._mPhones = [];
	for each (var phone in aXml.*::phones.*) {
		var type = _obmTypesMappings[_xS(phone.@label).replace("PREF;","")];
		if (type)
			this._mPhones[type] = _xS(phone.@number);
	}
	
	this._mEmails = [];
	for each (var mail in aXml.*::emails.*) {	
		var type = _obmTypesMappings[_xS(mail.@label).replace("PREF;","")];
		if (type)
			this._mEmails[type] = _xS(mail.@value);
	}
	
	this._mWebSites = [];
	for each (var site in aXml.*::websites.*) {
		var type = _obmTypesMappings[_xS(site.@label).replace("PREF;","")];
		if (type)
		this._mWebSites[type] = _xS(site.@url);
	}
	
	this._mAddresses = [];
	for each (var address in aXml.*::addresses.*) {
		var type = _obmTypesMappings[_xS(address.@label).replace("PREF;","")];
		if (type) {
			this._mAddresses[type] = {};
			this._mAddresses[type].street = _xS(address);
			this._mAddresses[type].state = _xS(address.@state);
			this._mAddresses[type].city = _xS(address.@town);
			this._mAddresses[type].country = _xS(address.@country);
			this._mAddresses[type].zipCode = _xS(address.@zip);
			this._mAddresses[type].expressPostal = _xS(address.@expressPostal);
		}
	}

}

ContactShell.prototype = {
	QueryInterface: function(aIID) {
		if (!aIID.equals(Ci.nsISupports) &&
			!aIID.equals(Ci.obmIRContact)) {
			throw Cr.NS_ERROR_NO_INTERFACE;
		}
		return this;
	},
	get uid() {return _xS(this._mXml.@uid)},
	get	addressbookid() {return _xL(this._mXml.*::addressbookid)},
	get	firstName() {return _xS(this._mXml.*::first)},
	get	lastName() {return _xS(this._mXml.*::last)},
	get	commonname() {return _xS(this._mXml.*::commonname)},
	get	aka() {return _xS(this._mXml.*::aka)},
	get	email() {return _sF(this._mEmails["email"])},
	get	email2() {return _sF(this._mEmails["email2"])},
	get	workPhone() {return _sF(this._mPhones["workPhone"])},
	get	homePhone() {return _sF(this._mPhones["homePhone"])},
	get	workFax() {return _sF(this._mPhones["workFax"])},
	get pagerPhone() {return _sF(this._mPhones["pagerPhone"])},
	get	mobilePhone() {return _sF(this._mPhones["mobilePhone"])},
	get	homeStreet() {return ((this._mAddresses["home"]) ? this._mAddresses["home"].street : null)},
	get	homeZipCode() {return ((this._mAddresses["home"]) ? this._mAddresses["home"].zipCode : null)},
	get	homeCity() {return ((this._mAddresses["home"]) ? this._mAddresses["home"].city : null)},
	get	homeState() {return ((this._mAddresses["home"]) ? this._mAddresses["home"].state : null)},
	get	homeExpressPostal() {return ((this._mAddresses["home"]) ? this._mAddresses["home"].expressPostal : null)},
	get	homeCountry() {return ((this._mAddresses["home"]) ? this._mAddresses["home"].country : null)},
	get homeWebPage() {return _sF(this._mWebSites["homeWebPage"])},
	get	title() {return _xS(this._mXml.*::title)},
	get	service() {return _xS(this._mXml.*::service)},
	get	company() {return _xS(this._mXml.*::company)},
	get	workStreet() {return ((this._mAddresses["work"]) ? this._mAddresses["work"].street : null)},
	get	workZipCode() {return ((this._mAddresses["work"]) ? this._mAddresses["work"].zipCode : null)},
	get	workCity() {return ((this._mAddresses["work"]) ? this._mAddresses["work"].city : null)},
	get	workState() {return ((this._mAddresses["work"]) ? this._mAddresses["work"].state : null)},
	get	workExpressPostal() {return ((this._mAddresses["work"]) ? this._mAddresses["work"].expressPostal :null)},
	get	workCountry() {return ((this._mAddresses["work"]) ? this._mAddresses["work"].country : null)},
	get workWebPage() {return _sF(this._mWebSites["workWebPage"])},
	get	comment() {return _xS(this._mXml.*::comment)},
	get caluri()  {return _xS(this._mXml.*::caluri)}
};

function AddressBookChangesResponseShell(aXml) {
	this._updated = [];
	this._removed = [];
	this._updatedAB = [];
	this._removedAB = [];
	this._lastSync = aXml.@lastSync.toString(); 

	for each (var updatedAB in aXml.*::addressbooks.*::updated.*) {
		this._updatedAB.push(new AddressBookShell(updatedAB));
	}

	for each (var removedAB in aXml.*::addressbooks.*::removed.*) {
		this._removedAB.push(removedAB.@uid.toString());
	}

	for each (var updated in aXml.*::contacts.*::updated.*) {
		this._updated.push(new ContactShell(updated));
	}
	
	for each (var removed in aXml.*::contacts.*::removed.*) {
		this._removed.push(removed.@uid.toString());
	}
}

AddressBookChangesResponseShell.prototype = {
	getUpdatedAddressBooks: function() {
		return this._updatedAB;
	},
	getRemovedAddressBooksIds: function() {
		return this._removedAB;
	},
	getUpdatedContacts: function() {
		return this._updated;
	},
	getRemovedContactIds: function() {
		return this._removed;
	},
	get lastsync() {
		return this._lastSync;
	}
};


// --------------- utilities objects to format data request

var _contactTypesMappings = {};
_contactTypesMappings["mobilePhone"] = "CELL;VOICE;X-OBM-Ref1";
_contactTypesMappings["pagerPhone"] = "PAGER;X-OBM-Ref1";
_contactTypesMappings["workPhone"] = "WORK;VOICE;X-OBM-Ref1";
_contactTypesMappings["workFax"] = "WORK;FAX;X-OBM-Ref1";
_contactTypesMappings["homePhone"] = "HOME;VOICE;X-OBM-Ref1";
_contactTypesMappings["email"] = "INTERNET;X-OBM-Ref1";
_contactTypesMappings["email2"] = "INTERNET;X-OBM-Ref2";
_contactTypesMappings["workWebPage"] = "URL;X-OBM-Ref1";
_contactTypesMappings["homeWebPage"] = "URL;X-OBM-Ref2";
_contactTypesMappings["home"] = "HOME;X-OBM-Ref1";
_contactTypesMappings["work"] = "WORK;X-OBM-Ref1";

// obmIRContact
function JSObjContactShell(aContact) {
	var obj = <contact/>;// xmlns="http://www.obm.org/xsd/sync/contact.xsd"/>;
	if (aContact.uid != null) {
		obj.@uid = aContact.uid;
	}
	
	_appendXMLElement(obj, "first", aContact.firstName);
	_appendXMLElement(obj, "last", aContact.lastName);
	_appendXMLElement(obj, "aka", aContact.aka);
	_appendXMLElement(obj, "commonname", aContact.commonname);
	_appendXMLElement(obj, "title", aContact.title);
	_appendXMLElement(obj, "service", aContact.service);
	_appendXMLElement(obj, "company", aContact.company);
	// FIXME is read-only ?
	_appendXMLElement(obj, "comment", aContact.comment);
	
	// emails
	obj.emails = <emails/>;
	if (aContact.email)
		obj.emails.mail += <mail label={_contactTypesMappings["email"]} value={aContact.email}/>;
	if (aContact.email2)
		obj.emails.mail += <mail label={_contactTypesMappings["email2"]} value={aContact.email2}/>;
		
	// phones
	obj.phones = <phones/>;
	if (aContact.workPhone)
		obj.phones.phone += <phone label={_contactTypesMappings["workPhone"]} number={aContact.workPhone}/>;
	if (aContact.homePhone)
		obj.phones.phone += <phone label={_contactTypesMappings["homePhone"]} number={aContact.homePhone}/>;
	if (aContact.workFax)
		obj.phones.phone += <phone label={_contactTypesMappings["workFax"]} number={aContact.workFax}/>;
	if (aContact.pagerPhone)
		obj.phones.phone += <phone label={_contactTypesMappings["pagerPhone"]} number={aContact.pagerPhone}/>;
	if (aContact.mobilePhone)
		obj.phones.phone += <phone label={_contactTypesMappings["mobilePhone"]} number={aContact.mobilePhone}/>;
		
	// home and word addresses
	obj.addresses = <addresses/>;
	if (aContact.homeStreet || aContact.homeCountry || aContact.homeCity || aContact.homeState || aContact.homeZipCode) {
		var home = <address label={_contactTypesMappings["home"]}>{_sX(aContact.homeStreet)}</address>;
		home.@country = _sX(aContact.homeCountry);
		home.@town = _sX(aContact.homeCity)
		home.@state = _sX(aContact.homeState);
		home.@zip = _sX(aContact.homeZipCode);
		//home.@expressPostal = _sX(aContact.homeExpressPostal);
		obj.addresses.address += home;
	}
	
	if (aContact.workStreet || aContact.workCountry || aContact.workCity || aContact.workState || aContact.workZipCode) {
		var work = <address label={_contactTypesMappings["work"]}>{_sX(aContact.workStreet)}</address>;
		work.@country = _sX(aContact.workCountry);
		work.@town = _sX(aContact.workCity);
		work.@state = _sX(aContact.workState);
		work.@zip = _sX(aContact.workZipCode);
		//work.@expressPostal = _sX(aContact.workExpressPostal);
		obj.addresses.address += work;
	}
	
	// websites
	obj.websites = <websites/>;
	if (aContact.homeWebPage)
		obj.websites.site += <site label={_contactTypesMappings["homeWebPage"]} url={aContact.homeWebPage}/>;
	if (aContact.workWebPage)
		obj.websites.site += <site label={_contactTypesMappings["workWebPage"]} url={aContact.workWebPage}/>;
	
	obj.instantmessaging = <instantmessaging/>;
	
	if (aContact.caluri)
		obj.caluri = <caluri>{_sX(aContact.caluri)}</caluri>;
	
	return ("<?xml version=\"1.0\" encoding=\"UTF-8\"?>" + obj.toString());
}
