/* ***** BEGIN LICENSE BLOCK *****
 *
 * Copyright (c) 1997-2008 Aliasource - Groupe LINAGORA
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of the
 *  License, (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  http://www.obm.org/
 *
 * ***** END LICENSE BLOCK ***** */

/*
This object is responsible of three things:
 - [A] to merge one (mergeOneCalendar()) or all calendars (mergeAllCalendars())
 - [B] to retrieve the list of calendars (getSubscribleCalendars()).
 - [C] to purge a calendar


For a merge:

1) Log in to the server
2) it retrieve the list of categories and stores the list of categories into the pref calendar.categories.names
3) 
	- it fills _localDiff with the list of event to add/update/remove into OBM
	- it fills _localDiff with the list of exceptions (FIXME: what is it ?) to add/update/
4) it prepares the list of operation to do on each calendars : == list of
		task which do a synchronisation with a specific calendar

   for each calendar
	- it fills _remoteDiff with informations from the response
	- if it is the first pass, it resolve conflicts (resolveConflicts)
		it changes _remoteDiff and _localDiff depending of conflicts found

5) when all sync data have been retrieved, _mergeRemoteDiff is called
	it processes all _remoteDiff updated/removed (remove or updateOrAdopt the events)

6) for the first pass, _sendLocalDiff is called. It calls OBM for each local modification
	the list is in _localDiff
	- calls removeEventByExtId/removeEvent for each removed events
	- calls _createEventObjFromId for each event to add or modify
		it converts the mozilla Event to an OBM event
		it calls CalProxy.createEvent or CalProxy.modifyEvent
		it update the change list into prefs

	_onSentLocalDiff is called for each OBM server call (after the http response)
		it update the local calendar
7) the second pass begins, it calls getSync for each calendar -> 5)
	FIXME: needs more documentation about this second pass (why there is a second pass?)
*/
function calendarSync() {
	// Init logger
	this.logger = Cc["@obm.org/calendar/logger;1"].createInstance(Ci.obmILogger);
	var writer = Cc["@obm.org/calendar/writer;1?type=unixfile"].getService(Ci.obmILoggerWriter);
	this.stringBundle = Cc["@mozilla.org/intl/stringbundle;1"].getService(Ci.nsIStringBundleService)
						  .createBundle("chrome://obm-extension/locale/calendar/errors_warnings.properties");
	this.logger.setLevelWriter(this.logger.LOAD_LEVEL, writer);
	this.logger.setHeader("calendarSync.js: ");
	this.logger.debug("new calendarSync()", false, false);

	this._timezone = calendarDefaultTimezone();
	this._localDiff = {};
	this._remoteDiff = {};

	// {calendar.id: { evtId: {action: "updated", ...} , ... }, ... }
	// the list of events that will be sent to the server
	this._evtsLocalDiff = {};

	this._supportEvtEx = false;
	this._supportExtId = false;
	this.session_id = null;
	this.syncVersion = null;
	this.login = ""; // filled by the loginTask
};

calendarSync.prototype = {

	//
	// Helpers
	//
	_getLocalizedMessage: function(aString) {
		try {
			return this.stringBundle.GetStringFromName("obm.syncerror." + aString);
		} catch(e) {
			return aString;
		}
	},

	initServerCapabilitiesFromSyncVersion: function() {
		if (UtilsOBM.compareVersions(this.syncVersion, "2.3.0") >= 0 ) {
			this._supportEvtEx = true;
		}

		//if (UtilsOBM.compareVersions(this.syncVersion, "2.3.11") >= 0 ) {
		this._supportExtId = true;
		// do a reset for extId <-> id changes
		if (!UtilsOBM.getBoolPref("extensions.obm.resetIdDone", true)) {
			return false;
		}
// 		}
		return true;
	},

	_getLastSync: function(calendar) {
		var calMgr = Cc["@mozilla.org/calendar/manager;1"].getService(Ci.calICalendarManager);
		var calName = calMgr.getCalendarPref_(calendar, "X-OBM-NAME");
		var lastSync = UtilsOBM.getCharPref("extensions.obm.changes." + Ci.calICalendar + "." + calName + "." + "lastSync", "0");
		// TODO : need an isValidDate func ?
		if (lastSync == "1970-01-01T00:00:00Z" || lastSync == "") {
			lastSync = "0";
		}
		return lastSync;
	},
	_setLastSync: function(calName, lastSync) {
		this.logger.debug("SETLASTSYNC (" + lastSync + ") for " + calName + ": " + "extensions.obm.changes." + Ci.calICalendar + "." + calName + "." + "lastSync=" + lastSync, false, false);
		try {
			UtilsOBM.setCharPref("extensions.obm.changes." + Ci.calICalendar + "." + calName + "." + "lastSync", lastSync);
		}
		catch (e) {
			// From time to time, the prefs.js is mysteriously locked and saving
			// the preferences fails with an NS_ERROR_FILE_ACCESS_DENIED error
			this.logger.error(e);
			// Complain to the user and ask her to try again
			var errorMsg  = this._getLocalizedMessage("prefSaveFailed");
			this.callerListener.onStop(-1, errorMsg);
			throw e;
		}
	},
	_mozEventToOBMEvent: function(aCalendar, calName, aEvent, aWillBeCreated) {
		var calMgr = Cc["@mozilla.org/calendar/manager;1"].getService(Ci.calICalendarManager);
		var localEvent = aEvent;
		var eventToSend = {
			recurrence: this
		};
		eventToSend.QueryInterface = function(aIID) {
			if (!aIID.equals(Ci.nsISupports)
				&& !aIID.equals(Ci.obmIREventRecurrence)
				&& !aIID.equals(Ci.obmIREvent)) {
				throw Cr.NS_ERROR_NO_INTERFACE;
			}
			return this;
		};
		if (!aWillBeCreated) {
			eventToSend.uid = UtilsOBM.getOBMEventId(this._supportExtId, localEvent);
		}
		if (this._supportExtId) {
			eventToSend.extId = localEvent.id;
		}
		if (localEvent.hasProperty('X-OBMEXT-IS-INTERNAL')) {
			eventToSend.isInternal = true;
		}

		if (localEvent.recurrenceId) {
			eventToSend.recurrenceId =  UtilsOBM.jsDateToRestDate(localEvent.recurrenceId.jsDate);
		}

		this.logger.debug("reading ganaration from localEvent : "+localEvent.generation, false, false);

		eventToSend.generation = UtilsOBM.getSequence(localEvent);
		eventToSend.owner = this.login.split("@")[0];
		eventToSend.ownerEmail= calMgr.getCalendarPref_(aCalendar, "X-OBM-EMAIL");

		var organizer = aEvent.organizer;
		//if (organizer != null && organizer.id != null && organizer.id != "") {
		//	eventToSend.ownerEmail = organizer.id.replace(/MAILTO:/i, "");
		//}
		var calPriority = localEvent.priority;

		if (calPriority >= 1 && calPriority <= 4) {
			eventToSend.priority = 3; // high priority in OBM
		} else if (calPriority == 5) {
			eventToSend.priority = 2; // medium priority in OBM
		} else if (calPriority >= 6 && calPriority <= 9) {
			eventToSend.priority = 1; // low priority in OBM
		} else {
			eventToSend.priority = 2; // medium priority in OBM, by default
		}

		switch (localEvent.privacy) {
		case "CONFIDENTIAL":
		case "PRIVATE":
			eventToSend.privacy = "1";
			break;
		case "PUBLIC":
		default:
			eventToSend.privacy = "0";
		}
		// 0 = public
		// 1 = private
		// ********************** //
		if (localEvent.startDate.isDate) {
			var begin = localEvent.startDate.jsDate;
			//begin.setHours(8);
			eventToSend.allday = true;
			eventToSend.date = UtilsOBM.jsDateToRestDate(begin);
			//if (localEvent.duration.inSeconds > 86400) {
			//  eventToSend.duration = localEvent.duration.inSeconds - 86400;
			//} else {
			eventToSend.duration = localEvent.duration.inSeconds;
			//}
		} else {
			eventToSend.allday = false;
			eventToSend.date = UtilsOBM.jsDateToRestDate(localEvent.startDate.jsDate);
			eventToSend.duration = localEvent.duration.inSeconds;
		}
		eventToSend.title = localEvent.title;
		eventToSend.description = localEvent.getProperty("DESCRIPTION");

		if (localEvent.getCategories) {
			// lightning 0.9 support
			var category = localEvent.getCategories({}).toString();
		} else {
			var category = localEvent.getProperty("CATEGORIES");
		}
		if (category) {
			var matches = category.match(/\[OBM\] (.*)/);
			if (matches) {
				eventToSend.category = matches[1];
			}
		}

		eventToSend.location = localEvent.getProperty("LOCATION");

		if (localEvent.getProperty("TRANSP") == "TRANSPARENT") {
			eventToSend.opacity = "TRANSPARENT";
		} else {
			eventToSend.opacity = "OPAQUE";
		}
		eventToSend.opacity = localEvent.getProperty("TRANSP");

		this._mozRecurrenceToOBMEvent(localEvent, eventToSend);

		var allAttendees = localEvent.getAttendees({});
		var goodAttendees = []; // list of attendees to send
		var subcribedEmail = calMgr.getCalendarPref_(aCalendar, "X-OBM-EMAIL");

		var organizerEmail = null;
		var organizerAdded = false; // indicate if the organizer has been found into the list of attendee
		var organizer = aEvent.organizer;
		if (organizer != null && organizer.id != null && organizer.id != "") {
			organizerEmail = organizer.id.replace(/MAILTO:/i, "");
		}

		var calAttendee = null; // the calAttendee object corresponding to the user
		for (var i = 0; i < allAttendees.length; i++) {
			if (allAttendees[i].participationStatus) {
				if (allAttendees[i].id.match("MAILTO:" + subcribedEmail, "i")) {
					calAttendee = allAttendees[i];
				}
				if (organizerEmail != null && allAttendees[i].id.match("MAILTO:" + organizerEmail, "i")) {
					organizerAdded = true;
				}
				goodAttendees.push(allAttendees[i]);
			}
		}
		var subscribedAdded = false;
		if (calAttendee) {
			subscribedAdded = true;
			if (!this._supportExtId) {
				// calAttendee object is immutable (read-only)
				// -> modifying a clone object
				var calAttendeeNew = calAttendee.clone();
				switch (localEvent.status) {
				case "CONFIRMED":
					calAttendeeNew.participationStatus = "ACCEPTED";
					break;
				case "CANCELLED":
					calAttendeeNew.participationStatus = "DECLINED";
					break;
				case "TENTATIVE":
					calAttendeeNew.participationStatus = "NEEDS-ACTION";
					break;
				}
				for (var i = 0; i < goodAttendees.length; i++) {
					if (goodAttendees[i].id == calAttendeeNew.id) {
						goodAttendees.splice(i, 1, calAttendeeNew);
						break;
					}
				}
			}
		}

		if (organizerEmail != null && !organizerAdded ) {
			goodAttendees.push(organizer.clone());
			if (organizerEmail == subcribedEmail) {
				subscribedAdded = true;
			}
		}

		eventToSend.attendees = [];
		for each(var calAttendee in goodAttendees) {
			var attendee = {};
			attendee.email = calAttendee.id.replace(/MAILTO:/i, "");
			attendee.displayName = calAttendee.commonName;
			attendee.status = this._getAcceptableParticipationState(calAttendee.participationStatus);
			switch (calAttendee.role) {
			case "REQ-PARTICIPANT":
				attendee.type = "REQ";
				break;
			case "OPT-PARTICIPANT":
			default:
				attendee.type = "OPT";
				break;
			}
			if (organizerEmail != null && calAttendee.id.match("MAILTO:" + organizerEmail, "i")) {
				attendee.isOrganizer = true;
			}
			else
				attendee.isOrganizer = false;
			eventToSend.attendees.push(attendee);
		}
		if (!subscribedAdded) {
			var attendee = {};
			attendee.email = subcribedEmail;
			attendee.displayName = subcribedEmail;
			attendee.status = "ACCEPTED";
			attendee.type = "REQ";
			if ( eventToSend.attendees.length == 0 ) {
				attendee.isOrganizer = true;
			}
			eventToSend.attendees.push(attendee);
		}

		eventToSend.getAttendees = function(count) {
			count.value = this.attendees.length;
			return this.attendees;
		};

		if (localEvent.alarmOffset === undefined) {
			var alarms = localEvent.getAlarms({});
			for each (var alarm in alarms) {
				if (alarm.related == Ci.calIAlarm.ALARM_RELATED_START) {
					var calAlert = alarm.offset;
					break;
				}
			}
		} else {
			calAlert = localEvent.alarmOffset;
		}

		var calAlertToXmlAlert = function(aCalAlert) {
			if (aCalAlert.inSeconds > 0) {
				return 0; //OBM do not support alert after events
			} else {
				return ( - 1 * aCalAlert.inSeconds);
			}
		};

		var obmAlert = -1;
		if (calAlert) {
			if (localEvent.alarmLastAck) {
				if (localEvent.recurrenceInfo) {
					obmAlert = calAlertToXmlAlert(calAlert);
				}
			} else {
				obmAlert = calAlertToXmlAlert(calAlert);
			}
		}
		eventToSend.alert = obmAlert;

		//recc exception events
		eventToSend.eventExceptions = [];
		eventToSend.getEventExceptions = function(count) {
			count.value = this.eventExceptions.length;
			return this.eventExceptions;
		};

		if (this._supportEvtEx) {
			if (localEvent.recurrenceInfo) {
				var eventExs = this._mozExceptionsToOBMEvents(localEvent.recurrenceInfo);
				for (var i=0; i < eventExs.length; i++) {
					this.logger.debug("mozEventEx recId:" + eventExs[i].recurrenceId, false, false);
					eventToSend.eventExceptions[i] = this._mozEventToOBMEvent(aCalendar, calName, eventExs[i], aWillBeCreated);
				}
			}
		}

		return eventToSend;
	},

	_mozRecurrenceToOBMEvent: function(aCalEvent, aObmEvent) {
		var recurrenceInfo = aCalEvent.recurrenceInfo;
		var exceptionIDs = [];
		var invalidChoice;
		var endDate = new Date();

		if (recurrenceInfo) {
			var recurrenceCount = recurrenceInfo.countRecurrenceItems();
			for (var i = 0; i < recurrenceCount; i++) {
				var recurrenceItem = recurrenceInfo.getRecurrenceItemAt(i);
				if (recurrenceItem.isNegative && recurrenceItem instanceof Ci.calIRecurrenceDate) {
					exceptionIDs.push(recurrenceItem.date);
				} else {
					if (!recurrenceItem.isNegative && recurrenceItem instanceof Ci.calIRecurrenceRule) {

						// !isFinite ==> repeat forever
						// -> transform to "until event start date + 1 year"
						if (!recurrenceItem.isFinite) {
							endDate = aCalEvent.startDate.jsDate;
							endDate.setHours(23);
							endDate.setMinutes(59);
							endDate.setSeconds(59);
							endDate.setMilliseconds(999);
							endDate.setFullYear(endDate.getFullYear() + 1); // add a year
							// correct timezone offset
							endDate = new Date(endDate.getTime() - endDate.getTimezoneOffset() * 60 * 1000);
						} else {
							if (recurrenceItem.isByCount) {
								// end date ==> date of last occurrence
								var previousOccurence;
								var nextOccurrence = aCalEvent;

								do {
									previousOccurence = nextOccurrence;
									var occurrenceTime = previousOccurence.startDate;
									nextOccurrence = recurrenceInfo.getNextOccurrence(occurrenceTime);
								} while ( nextOccurrence != null );

								try {
									endDate = previousOccurence.endDate.jsDate;
								} catch (e) {
									endDate = previousOccurence.untilDate.jsDate;
								}
							} else { // recurrence lasts up to a certain date
								try {
									endDate = recurrenceItem.endDate.jsDate;
								} catch (e) {
									endDate = recurrenceItem.untilDate.jsDate;
								}
							}
						}

						switch (recurrenceItem.type) {
						case "DAILY":

							var onDays = recurrenceItem.getComponent("BYDAY", {});
							if (onDays.length > 0) {
								//on every weekday
								var days = ['0', '0', '0', '0', '0', '0', '0'];
								for (var j = 0; j < onDays.length; j++) {
									days[onDays[j] - 1] = '1';
								}
								days = days.join("");

								aObmEvent.kind = "weekly";
								aObmEvent.frequence = recurrenceItem.interval;
								aObmEvent.days = days;
								break;
							}
							aObmEvent.kind = "daily";
							aObmEvent.frequence = recurrenceItem.interval;
							aObmEvent.days = "";
							break;

						case "WEEKLY":
							aObmEvent.kind = "weekly";
							aObmEvent.frequence = recurrenceItem.interval;
							var onDays = recurrenceItem.getComponent("BYDAY", {});
							var days = ['0', '0', '0', '0', '0', '0', '0'];
							if (!onDays || onDays === null || onDays.lengh === 0 || onDays == "") {
								// lightning bug : no day selected
								// -> no day stored but start day of event used in views
								days[aCalEvent.startDate.weekday] = '1';
							} else {
								for (var j = 0; j < onDays.length; j++) {
									days[onDays[j] - 1] = '1';
								}
							}
							days = days.join("");
							// make Sunday day 7 no more needed in obm 2.2
							//days = days.replace(/(0|1)([0|1]{6})/, "$2$1");
							aObmEvent.days = days;
							break;

						case "MONTHLY":
							var onDays = recurrenceItem.getComponent("BYDAY", {});
							//dump("recc MONTHLY, onDays" + onDays[0] + "\n");
							if (onDays.length == 1) {
								aObmEvent.kind = "monthlybyday";
								aObmEvent.frequence = recurrenceItem.interval;
								aObmEvent.days = "";
								break;
							}
							onDays = recurrenceItem.getComponent("BYMONTHDAY", {});
							//dump("recc, onDays" + onDays[0] + "\n");
							if (onDays.length == 1) {
								aObmEvent.kind = "monthlybydate";
								aObmEvent.frequence = recurrenceItem.interval;
								aObmEvent.days = "";
								break;
							}
							// lightning bug : no day or date selected
							// force recc by date
							aObmEvent.kind = "monthlybydate";
							aObmEvent.frequence = recurrenceItem.interval;
							aObmEvent.days = "";
							//invalidChoice = true;
							break;
						case "YEARLY":
							aObmEvent.kind = "yearly";
							aObmEvent.frequence = recurrenceItem.interval;
							aObmEvent.days = "";
							break;
						default:
							invalidChoice = true;
							break;
						}
					}
				}
			}
		}
		aObmEvent.exceptions = [];
		aObmEvent.getExceptions = function(count) {
			count.value = this.exceptions.length;
			return this.exceptions;
		};

		if (!recurrenceInfo || invalidChoice) {
			aObmEvent.kind = "none";
			aObmEvent.frequence = 1;
			aObmEvent.days = "";
		} else {
			if (!this._supportEvtEx) {
				exceptionIDs = exceptionIDs.concat(recurrenceInfo.getExceptionIds({}));
			}
			for each(var exceptionID in exceptionIDs) {
				aObmEvent.exceptions.push(UtilsOBM.jsDateToRestDate(exceptionID.jsDate).toString());
			}
		}
		if (recurrenceItem && !recurrenceItem.isFinite) {
			aObmEvent.end = null;
		} else {
			aObmEvent.end = UtilsOBM.jsDateToRestDate(endDate);
		}

		// OBM can't handle main event if it is not in recurrence (Bug #171)
		// So event is set to be on the first occurrence
		if (recurrenceInfo) {
			var aDate = aCalEvent.startDate.clone();
			aDate.resetTo(aDate.year, aDate.month, aDate.day, aDate.hour, aDate.minute, aDate.second - 1, aDate.timezone);
			var firstRecurrenceItem = recurrenceInfo.getNextOccurrence(aDate);
			if (firstRecurrenceItem && aCalEvent.startDate.compare(firstRecurrenceItem.startDate)) {
				this.logger.debug("First occurrence of the event (" + firstRecurrenceItem.startDate + ") is not the start date (" + aCalEvent.startDate + "). Start date is set to first occurence.", false, false);
				aObmEvent.date = UtilsOBM.jsDateToRestDate(firstRecurrenceItem.startDate.jsDate);
			}
		}
	},

	_calGetItem: function(calendar, remoteEvent, callback) {
		var self = this;

		var listener = {
			localEvent : null
		};

		listener.onOperationComplete = function(aCalendar, aStatus, aOperationType, aId, aDetail) {
			//<TO_REMOVE>
			// Fix second onOperationComplete call (bug LT 0.8)
			// Fixed in LT 0.9
			try {
				if (aDetail.toString() == "Can't deduce item type based on QI") {
					dump("_calGetItem, aDetail = Can't deduce item type based on QI\n");
					return;
				}
			} catch(e) {
				dump("_calGetItem, aDetail: not an object?\n");
				dump(e);
			}
			//</TO_REMOVE>
			//dump("--aStatus: " + aStatus + " aOperationType: " + aOperationType + " aId: " + aId + " aDetail: " + aDetail + "\n");
			// FIXME : bug lightning ? aStatus for getItem is allways 0
			if (!Components.isSuccessCode(aStatus) || !this.localEvent) {
				// nothing to do ?
				callback(null);
			} else {
				var localEvent;
				try {
					localEvent = aDetail.QueryInterface(Ci.calIEvent)
				} catch(e) {
					localEvent = this.localEvent;
				}

				callback (localEvent);
			}
		};
		listener.onGetResult = function(aCalendar, aStatus, aItemType, aDetail, aCount, aItems) {
			if (aCount > 0) {
				this.localEvent = aItems[0];
			}
		};
		if (this._supportExtId)
			calendar.getItem(remoteEvent.extId, listener);
		else
			calendar.getItem(remoteEvent.uid, listener);
	},

	// --
	// this._localDiff[calendar.id] contains local changes
	// this._remoteDiff[calName] contains remote changes
	//
	resolveConflicts: function(calName) {
		this.logger.debug("resolveConflicts("+calName+")", false, false);

		var self = this;

		// Only work on OBM calendars
		var serverPriority = (UtilsOBM.getCharPref("extensions.obm.override", "server") == "server");
		var calMgr = Cc["@mozilla.org/calendar/manager;1"].getService(Ci.calICalendarManager);
		var calendar;
		for (var i = 0; i < this.calendars.length; i++) {
			calendar = this.calendars[i];
			var name = calMgr.getCalendarPref_(calendar, "X-OBM-NAME");
			if (name == calName) {
				break;
			}
		}

		if ( !this._localDiff[calendar.id] ) {
			this._localDiff[calendar.id] = {event:{}};
		}

		var localDiff = this._localDiff[calendar.id];
		var changesList = this._remoteDiff[calName];

		this.logger.debug("  local events : "+ JSON.stringify(localDiff),false,false);

		var conflicts = [];
		var localToSend = {};
		var remoteToApply = {participationChanged: [], updated: [], removed: [], lastSync : this._remoteDiff[calName].lastSync };

		var getConflict = function (id) {
			for (var i in conflicts ) {
				if ( conflicts[i].id == id ) {
					return conflicts[i];
				}
			}
			return false;
		}

		var getRemoteChange = function(type,id) {
			if ( type == "updated" ) {
				for ( var i in changesList.updated ) {
					if ( changesList.updated[i].extId == id ) {
						return changesList.updated[i];
					}
				}
			} else if ( type == "removed" ) {
				for ( var i in changesList.removed ) {
					if (changesList.removed[i] == id ) {
						return id;
					}
				}
			} else if ( type == "participationChanged" ) {
				for ( var i in changesList.participationChanged ) {
					if (changesList.participationChanged[i].extId == id ) {
						return changesList.participationChanged[i];
					}
				}
			}
			return null;
		};

		if (this.logger.getLevelWriter(this.logger.DEBUG_LEVEL)) {
			this.logger.debug("resolveConflicts() localDiff: "+JSON.stringify(localDiff), false, false);

			for ( var type in this._remoteDiff[calName] ) {
				this.logger.debug("resolveConflicts() remoteDiff["+type+"] length: "+changesList[type].length, false, false);
			}
		}

		// find conflicts
		for ( var i in changesList.updated ) {
			var id = changesList.updated[i].extId;
			this.logger.debug("resolveConflicts() remoteDiff.updated: "+changesList.updated[i].extId, false, false);
			if ( localDiff.event[id] && localDiff.event[id].action ) {
				conflicts.push({id: id, local: localDiff.event[id].action, remote: "updated" });
			} else {
				remoteToApply.updated.push(changesList.updated[i]);
			}
		}
		for ( var i in changesList.participationChanged ) {
			var id = changesList.participationChanged[i].extId;
			this.logger.debug("resolveConflicts() remoteDiff.participationChanged: "+changesList.participationChanged[i].extId, false, false);
			if ( localDiff.event[id] && localDiff.event[id].action ) {
				conflicts.push({id: id, local: localDiff.event[id].action, remote: "participationChanged" });
			} else {
				remoteToApply.participationChanged.push(changesList.participationChanged[i]);
			}
		}
		for ( var i in changesList.removed ) {
			var id = changesList.removed[i];
			this.logger.debug("resolveConflicts() remoteDiff.removed: "+id, false, false);
			if ( localDiff.event[id] && localDiff.event[id].action ) {
				conflicts.push({id: id, local: localDiff.event[id].action, remote: "removed" });
			} else {
				remoteToApply.removed.push(changesList.removed[i]);
			}
		}

		// get local event without conflicts
		for ( var i in localDiff.event ) {
			if ( !getConflict(i) ) {
				localToSend[i] = localDiff.event[i];
			}
		}

		this.logger.debug("resolveConflicts() got "+conflicts.length+" conflicts",false,false);
		this.logger.debug("   details: "+JSON.stringify(conflicts),false,false);

		for ( var i in conflicts ) {
			var conflict = conflicts[i];

			if ( conflict.local == "removed" ) {
				if ( conflict.remote == "removed" ) {
					// event flagged as 'removed from local DB' AND "remove from remote DB"
					// So : nothing to do.
					remoteToApply.removed.push(conflict.id);
				} else if ( conflict.remote == "updated" ) {
					// local event is flagged as deleted, remote event as updated
					// check sequence
					var localSequence = localDiff.event[conflict.id].generation || 0;
					var remoteEvt = getRemoteChange.call(this,conflict.remote, conflict.id);
					var remoteSequence = remoteEvt.generation;
					if ( remoteSequence > localSequence ) {
						remoteToApply.updated.push(remoteEvt);
					} else if ( remoteSequence < localSequence ) {
						localToSend[conflict.id] = localDiff.event[conflict.id];
					} else {
						// sequence are the same : fallback to synchro preferences
						if (serverPriority) {
							remoteToApply.updated.push(remoteEvt);
						} else {
							localToSend[conflict.id] = localDiff.event[conflict.id];
						}
					}
				} else if ( conflict.remote == "participationChanged") {
					// ignore participation change as we remove the event
					localToSend[conflict.id] = localDiff.event[conflict.id];
				}
			} else if ( conflict.local == "updated") {
				if ( conflict.remote == "removed" ) {
					// don't have sequence from server on delete events... :-(
					if (serverPriority) {
						remoteToApply.removed.push(conflict.id);
					} else {
						localToSend[conflict.id] = localDiff.event[conflict.id];
					}
				} else if ( conflict.remote == "updated" ) {
					var localSequence = localDiff.event[conflict.id].generation || 0;
					var remoteEvt = getRemoteChange(conflict.remote, conflict.id);
					var remoteSequence = remoteEvt.generation;
					if ( remoteSequence > localSequence ) {
						remoteToApply.updated.push(remoteEvt);
					} else if ( remoteSequence < localSequence ) {
						localToSend[conflict.id] = localDiff.event[conflict.id];
					} else {
						// sequence are the same : fallback to synchro preferences
						if (serverPriority) {
							remoteToApply.updated.push(remoteEvt);
						} else {
							localToSend[conflict.id] = localDiff.event[conflict.id];
						}
					}
				} else if ( conflict.remote == "participationChanged") {
					// event is updated on local, and participationChanged on remote side...
					// => update local event with remote participation change and send the updated event
					this._updateParticipation(calendar, getRemoteChange(conflict.remote, conflict.id));
					localToSend[conflict.id] = localDiff.event[conflict.id];
				}
			} else if ( conflict.local == "added" ) {
				// hardcore: we created an event locally and it is already updated on remote side...
				var remoteEvt = getRemoteChange(conflict.remote, conflict.id);
				if ( remoteEvt.owner == this.login.split("@")[0]) {
					//created in this tb, accepted by an attendee in other tb, created in obm by attendee sync :
					//do not send update or attendee status will be lost
				} else {
					//created in Ui, accepeted by mail before sync in this tb : add to updated to send participation
					localToSend[conflict.id] = localDiff.event[conflict.id];
				}

			} else if ( conflict.local == "participationChanged") {

				if ( conflict.remote == "participationChanged" ) {
					// bc the remote event changed we ignore the local participationstate change
					var remoteEvt = getRemoteChange(conflict.remote, conflict.id);
					remoteToApply[conflict.remote].push(remoteEvt);
				} else if ( conflict.remote == "updated" ) {
					// event has been participation changed locally and updated remotely
					var localEmail = calMgr.getCalendarPref_(calendar, "X-OBM-EMAIL");
					var remoteEvt = getRemoteChange(conflict.remote, conflict.id);

					this._calGetItem(calendar, remoteEvt, function (localEvent) {
						if (!localEvent) {
							// this case should not appear
							self.logger.debug("_resolveConflicts: LocalEvent not found", false, false);
							return;
						} else {
							self.logger.debug("_resolveConflicts: localEvent.id= " + localEvent.id, false, false);
						}
						
						var remoteSequence = remoteEvt.generation;
						var localSequence = UtilsOBM.getSequence(localEvent);
						
						self.logger.debug("_resolveConflicts: localEvent sequence = " + localSequence + ", remoteEvent sequence = " + remoteSequence , false, false);
						if ( localSequence >= remoteSequence ) {
							// if localSequence = remoteSequence, we accepted/refused an event via the blue bar before syncing the calendar
							// retrieve the attendee object from the localEvent, to retrieve its status
							var localAttendee = localEvent.getAttendeeById("MAILTO:" + localEmail);
							self.logger.debug("_resolveConflicts: localAttendee.id= " + localAttendee.id, false, false);
							var participationStatus = localAttendee.participationStatus;
							self.logger.debug("_resolveConflicts: localAttendee.participationStatus= " + participationStatus, false, false);
							// set the participation state
							// of the local attendee in the remote event (bc we will record the remote evt in the local calendar on next step)
							// now retrieve the attendee object from the remoteEvt
							var remoteAttendee = remoteEvt.getAttendeeByEmail(localEmail);
							self.logger.debug("_resolveConflicts: remoteAttendee= " + remoteAttendee.email + ", remoteAttendee participation Status = "+ remoteAttendee.status, false, false);
							if ( remoteAttendee && remoteAttendee.status != participationStatus ) {
								self.logger.debug("_resolveConflicts: update remote attendee's participation", false, false);
								// we modify the participationStatus into the remoteAttendee
								remoteAttendee.status = participationStatus;
								// we send to the server our participationChange
								localToSend[conflict.id] = localDiff.event[conflict.id];
							}
						} else {
							// the remote event has been updated since our acceptance/declination, we forget our participationState and apply the remote event
							remoteToApply[conflict.remote].push(remoteEvt);
						}
												
					});

					

				} else if ( conflict.remote == "removed") {
					// the event has been removed remotely : we don't care about participation change
					remoteToApply.removed.push(conflict.id);
				}
			}
		}
		//this.logger.debug("resolveConflicts() remoteDiff.removed: "+JSON.stringify(this._remoteDiff[calName]).removed, false, false);
		//this.logger.debug("resolveConflicts() remoteDiff.participationChanged: "+JSON.stringify(this._remoteDiff[calName]).participationChanged, false, false);
		//this.logger.debug("resolveConflicts() remoteDiff.updated: "+JSON.stringify(this._remoteDiff[calName]).updated, false, false);

		this._remoteDiff[calName] = remoteToApply;
		this._localDiff[calendar.id].event = localToSend;
		this.logger.debug("end of resolveConflicts()", false, false);
	},
	/**
	 * return the list of subscribed OBM Calendar
	 * @return array  array of calICalendar ?
	 */
	getSubscribedOBMCalendar: function() {
		var calMgr = Cc["@mozilla.org/calendar/manager;1"].getService(Ci.calICalendarManager);
		var calendars = calMgr.getCalendars({});
		var res = new Array();
		for each(var calendar in calendars) {
			if (calendar.type == "obm") {
				res.push(calendar);
			}
		}
		return res;
	},

	/**
	 * the method which begins the real synchronisation, after the login process
	 * and the retrieve of the categories list
	 * (and the retrieve of list of OBM calendar if it is not already loaded)
	 */
	loadLocalDiff: function(calendars) {
		this.logger.debug("loadLocalDiff()", false, false);
		var calMgr = Cc["@mozilla.org/calendar/manager;1"].getService(Ci.calICalendarManager);
		this.callerListener.onNewOperation("getLocalDIff", null);
		this.calendars = calendars;

		for (var i = 0; i < this.calendars.length; i++) {

			var calName = calMgr.getCalendarPref_(this.calendars[i], "X-OBM-NAME");
			var changeTypes = ["added", "updated", "removed", "participationChanged"];
			this._localDiff[this.calendars[i].id] = {};
			this._localDiff[this.calendars[i].id]["event"] = {};
			this._localDiff[this.calendars[i].id]["task"] = {};

			this.logger.debug("_getLocalDiff : parsing "+calName,false,false);

			var allEventsChanges = UtilsOBM.getAllCalendarSyncItems(this.calendars[i]);

			this.logger.debug("_getLocalDiff : allEventsChanges : "+JSON.stringify(allEventsChanges),false,false);
/** this thing produces events duplication on reccurence exceptions
			for ( var j in allEventsChanges ) {
				if ( allEventsChanges[j].action && ( allEventsChanges[j].action == "added" || allEventsChanges[j].action == "updated" ) ) {
					//Does this thing work ???
					// WARNING _exceptionToItem calls asynchronous method of calICalendar objects
						// FIXME we should implement callbacks here... ?
					var addedIds = this._exceptionToItem(this.calendars[i], j);
					for each(var id in addedIds) {
						allEventsChanges[id] = {action: "added"};
					}
				}
			}
*/
			if (this.logger.getLevelWriter(this.logger.DEBUG_LEVEL)) {
				this.logger.debug("_getLocalDiff: ", false, false);
				for ( var j in allEventsChanges ) {
					this.logger.debug("   " + allEventsChanges[j].action+ " "+j, false, false);
				}
			}
			this._localDiff[this.calendars[i].id]["event"] = allEventsChanges;
			/*
			this.logger.debug("_getLocalDiff updated: " + this._localDiff[calName].event["updated"], false, false);
			this.logger.debug("_getLocalDiff removed: " + this._localDiff[calName].event["removed"], false, false);
			this.logger.debug("_getLocalDiff added: " + this._localDiff[calName].event["added"], false, false);
			this.logger.debug("_getLocalDiff participationChanged: " + this._localDiff[calName].event["participationChanged"], false, false);
			*/
		}
		this.callerListener.onOperationDone(true, null);
	},
	_mozExceptionsToOBMEvents: function(aReccurenceInfo) {
		var aExceptionIDs = aReccurenceInfo.getExceptionIds({});
		var exceptionEvents = [];
		if (aExceptionIDs.length != 0) {

			for each(var aExceptionID in aExceptionIDs) {
				var ocurrence = aReccurenceInfo.getOccurrenceFor(aExceptionID, false);
				this.logger.debug("exception ocurrence:" + aExceptionID , false, false);
				exceptionEvents.push(ocurrence);
			}
		}
		return exceptionEvents;
	},
	_exceptionToItem: function(calendar, aId) {

		this.logger.debug("_exceptionToItem()", false, false);
		this.logger.debug("  calendar: " + calendar, false, false);
		this.logger.debug("  aId: " + aId, false, false);
		var listener = {};
		this.addedIds = [];
		var self = this;
		listener.calendar = calendar;
		listener.onOperationComplete = function(aCalendar, aStatus, aOperationType, aId, aDetail) {
			// <TO_REMOVE>
			// Fix second onOperationComplete call (bug LT 0.8)
			// Fixed in LT 0.9
			try {
				if (aDetail.toString() == "Can't deduce item type based on QI") {
					return;
				}
			} catch(e) {}
			// </TO_REMOVE>
			if (Components.isSuccessCode(aStatus)) {
				var calMgr = Cc["@mozilla.org/calendar/manager;1"].getService(Ci.calICalendarManager);
				var calName = calMgr.getCalendarPref_(this.calendar, "X-OBM-NAME");
				var localEvent = this.localEvent;

				//get exception
				try {
					if (localEvent.recurrenceInfo != null) {
						var aExceptionIDs = localEvent.recurrenceInfo.getExceptionIds({});
						// has Exception
						if (aExceptionIDs.length != 0) {

							for each(var aExceptionID in aExceptionIDs) {
								var ocurrence = localEvent.recurrenceInfo.getOccurrenceFor(aExceptionID, false);

								//var calMgr = Cc["@mozilla.org/calendar/manager;1"].getService(Ci.calICalendarManager);
								var calEvent = Cc["@mozilla.org/calendar/event;1"].createInstance(Ci.calIEvent);
								var uuidgen = Components.classes["@mozilla.org/uuid-generator;1"].getService(Components.interfaces.nsIUUIDGenerator);
								var uuid = uuidgen.generateUUID().toString();

								calEvent.calendar = this.calendar;
								calEvent.id = uuid.substring(1, uuid.length - 1);
								calEvent.title = ocurrence.title;
								calEvent.setProperty("DESCRIPTION", ocurrence.getProperty("DESCRIPTION"));
								calEvent.setProperty("LOCATION", ocurrence.getProperty("LOCATION"));
								calEvent.startDate = ocurrence.startDate;
								calEvent.endDate = ocurrence.endDate;
								calEvent.setProperty("CATEGORIES", ocurrence.getProperty("CATEGORIES"));
								calEvent.priority = ocurrence.priority;
								calEvent.privacy = ocurrence.privacy;
								calEvent.setProperty("TRANSP", localEvent.getProperty("TRANSP"));

								var attendees = ocurrence.getAttendees({});
								for each(var attendee in attendees) {
									calEvent.addAttendee(attendee);
								}

								var listenerAdd = {};
								var caller = this;
								listenerAdd.pcaller = self;
								listenerAdd.onOperationComplete = function(cal, res, op, id, item) {
									this.pcaller.addedIds.push(id);
								};
								this.calendar.adoptItem(calEvent, listenerAdd);
							} //end for
						}
					}
				} catch(e) {
					self.logger.debug("error getting exception items: " + e, false, false);
				}
			}
		};
		listener.onGetResult = function(aCalendar, aStatus, aItemType, aDetail, aCount, aItems) {
			if (aCount > 0) {
				this.localEvent = aItems[0];
			}
		};

		// WARNING this method is asynchronous, and here, we assume that all things
		// append synchronously !!!!
		calendar.getItem(aId, listener);
		return this.addedIds;
	},

	_getCalendarEmails: function() {
		// Returns the names of the current OBM calendars
		var calMgr = Cc["@mozilla.org/calendar/manager;1"].getService(Ci.calICalendarManager);
		var calEmails = [];
		var calendars = this.getSubscribedOBMCalendar();
		for (var i = 0; i < calendars.length; i++) {
			var cal = calendars[i];
			var calEmail = calMgr.getCalendarPref_(cal, "X-OBM-EMAIL");
			calEmails.push(calEmail);
		}
		return calEmails;
	},

	/**
	 * @param CalendarArrayShell calendarInfos
	 */
	processCalendarMetadata: function(calendarInfos) {
		var calendarManager = UtilsOBM.getCalendarManager();
		var localCalendars = UtilsOBM.getSubscribedOBMCalendar();
		var localCalendarsHash = this._getCalendarsAsHash(localCalendars);
		for (var i = 0 ; i < calendarInfos.length ; i++) {
			var remoteCalInfo = calendarInfos.getElementAt(i);
			var localCalendar = localCalendarsHash[remoteCalInfo.mail];
			if (!localCalendar)
				throw "No calendar found with OBM email " + remoteCalInfo.mail;
			var remoteCalHasReadRights = remoteCalInfo.read;
			var remoteCalHasWriteRights = remoteCalInfo.write;
			var localCalHasWriteRights = (! localCalendar.readOnly);

			// The read right has been removed in OBM UI, remove the calendar
			// altogether
			if (!remoteCalHasReadRights) {
				calendarManager.unregisterCalendar(localCalendar);
				calendarManager.deleteCalendar(localCalendar);
			}
			else if (remoteCalHasWriteRights != localCalHasWriteRights) {
				// The rights to write events have been disabled remotely
				// Discard any changes done in this calendar since last time

				// Remove the calendar, and re-add it
				var oldColor = localCalendar.getProperty("color");
				calendarManager.unregisterCalendar(localCalendar);
				calendarManager.deleteCalendar(localCalendar);
				var newCalendar = this._subscribeTo(remoteCalInfo);
				// _subscribeTo has assigned a random color to the 'new'
				// calendar. We need to assign the old color and re-add it
				newCalendar.setProperty("color", oldColor);
				var wasWriteRightsRemoved = !remoteCalHasWriteRights;
				// Don't attempt to sync events, since we don't have the rights
				// to write on this calendar anymore
				if (wasWriteRightsRemoved)
					UtilsOBM.removeAllCalendarSyncItems(localCalendar);
			}
		}
	},

	storeRemoteDiff : function (calendar, calName, syncResponse) {
		this._remoteDiff[calName] = {};
		this._remoteDiff[calName]["updated"] = syncResponse.getUpdatedEvents(); // list of EventShell
		this._remoteDiff[calName]["participationChanged"] = syncResponse.getParticipationChanges(); // list of ParticipationChangeShell
		var lastSync = this._getLastSync(calendar);
		if (lastSync != "0") {
			if (this._supportExtId) {
				this._remoteDiff[calName]["removed"] = syncResponse.getRemovedEventExtIds(); // list of ids
			} else {
				this._remoteDiff[calName]["removed"] = syncResponse.getRemovedEventIds(); // list of ids
			}
		} else {
			this._remoteDiff[calName]["removed"] = [];
		}
		this._remoteDiff[calName].lastSync = syncResponse.lastsync;
	},

	/**
 * Traduit une récurrence du format OBM au format Calendar.
 *
 * @param {obmIREventRecurrence}
 *            aRecurrence
 * @pram {calIEvent} aCalEvent L'évènement auquel la récurrence est associée
 * @type calIRecurrence
 */
	_obmRecurrenceToCalRecurrence: function(aRecurrence, aCalEvent) {
		var recurrenceInfo = Cc["@mozilla.org/calendar/recurrence-info;1"].createInstance(Ci.calIRecurrenceInfo);
		recurrenceInfo.item = aCalEvent;

		var recRule = Cc["@mozilla.org/calendar/recurrence-rule;1"].createInstance(Ci.calIRecurrenceRule);

		switch (aRecurrence.kind) {
		case "daily":
			recRule.type = "DAILY";
			recRule.interval = aRecurrence.frequence;
			break;

		case "weekly":
			recRule.type = "WEEKLY";
			recRule.interval = aRecurrence.frequence;

			var onDays = [];

			var onDaysOBM = aRecurrence.days.split("");
			//onDaysOBM.unshift(onDaysOBM.pop());
			for (var i = 0; i < onDaysOBM.length; i++) {
				if (onDaysOBM[i] == "1") {
					onDays.push((i + 1).toString());
				}
			}

			if (onDays.length > 0) {
				recRule.setComponent("BYDAY", onDays.length, onDays);
			}
			break;

		case "monthlybyday":
			recRule.type = "MONTHLY";
			recRule.interval = aRecurrence.frequence;
			recRule.setComponent("BYDAY", 1, [Math.round(aCalEvent.startDate.day / 7) * 8 + aCalEvent.startDate.weekday + 1]);
			break;

		case "monthlybydate":
			recRule.type = "MONTHLY";
			recRule.interval = aRecurrence.frequence;
			recRule.setComponent("BYMONTHDAY", 1, [aCalEvent.startDate.day]);
			break;

		case "yearly":
			recRule.type = "YEARLY";
			recRule.interval = aRecurrence.frequence;
			break;

		case "none":
		default:
			return null;
			break;
		}

		if (!aRecurrence.end || aRecurrence.end == "") {
			recurrenceInfo.isFinite = false;
		} else {
			var endDate = jsDateToDateTime(UtilsOBM.restDateToJSDate(aRecurrence.end)).getInTimezone(this._timezone);
			endDate.timezone = this._timezone;
			if (!aCalEvent.startDate.isDate) {
				endDate.hour = aCalEvent.startDate.hour;
				endDate.minute = aCalEvent.startDate.minute;
				endDate.second = aCalEvent.startDate.second;
			} else {
				endDate.isDate = true;
			}
			try {
				recRule.endDate = endDate;
			} catch (e) {
				recRule.untilDate = endDate;
			}
		}
		recurrenceInfo.insertRecurrenceItemAt(recRule, 0);

		// first, remove all current exception IDs
		var exceptionIDs = recurrenceInfo.getExceptionIds({});
		for each(var exceptionID in exceptionIDs) {
			recurrenceInfo.removeExceptionFor(exceptionID);
		}

		// now, add all "new" exception IDs
		var exceptions = aRecurrence.getExceptions({});
		for each(var exception in exceptions) {
			var exceptionID = jsDateToDateTime(UtilsOBM.restDateToJSDate(exception)).getInTimezone(this._timezone);
			exceptionID.timezone = this._timezone;

			if (aCalEvent.startDate.isDate) { // event is all day
				exceptionID.isDate = true;
			}

			try {
				recurrenceInfo.removeOccurrenceAt(exceptionID); // getExceptionFor(exceptionID,
				// true /* aCreate */);
			} catch(e) {
				// in case exception date isn't a valid occurence date of the recurrence
				// date set
			}
		}

		// add eventExceptions
		if (this._supportEvtEx) {
			var eventExceptions = aRecurrence.getEventExceptions({}) || [];
			this.logger.debug("eventExceptions:" + eventExceptions.length, false, false);
			for each(var eventException in eventExceptions) {
				try {
					recurrenceInfo.modifyException(this._createLocalEvent(eventException, aCalEvent.calendar, aCalEvent), false);
				} catch (e) {
					this.logger.error(e, true, false);
				}
			}
		}
		return recurrenceInfo;
	},

	_createLocalEvent: function(aEvent /* obmIREvent */, calendar, aParentItem) {
		this.logger.debug("_createLocalEvent()", false, false);
		this.logger.debug("  event:", false, false);
		this.logger.debug("    - title: " + aEvent.title, false, false);
		this.logger.debug("    - uid:   " + aEvent.uid, false, false);
		this.logger.debug("    - extId: " + aEvent.extId, false, false);
		this.logger.debug("    - generation: " + aEvent.generation, false, false);

		if (aParentItem) {
			var calEvent = aParentItem.clone();
		} else {
			var calEvent = Cc["@mozilla.org/calendar/event;1"].createInstance(Ci.calIEvent);
		}
		calEvent.calendar = calendar;
		if (aParentItem) {
			calEvent.parentItem = aParentItem;
		} else {
			if (aEvent.uid) UtilsOBM.setOBMEventId(this._supportExtId, calEvent, aEvent.uid);
			if (this._supportExtId) {
				if (aEvent.extId) calEvent.id = aEvent.extId;
			}
		}

		calEvent.title = aEvent.title;
		UtilsOBM.setSequence(calEvent,aEvent.generation);
		if (aEvent.description != "") {
			calEvent.setProperty("DESCRIPTION", aEvent.description);
		}
		if (aEvent.location != "") {
			calEvent.setProperty("LOCATION", aEvent.location);
		}
		if (aEvent.opacity != "") {
			calEvent.setProperty("TRANSP", aEvent.opacity);
		}
		var jsStartDateTime = jsDateToDateTime(UtilsOBM.restDateToJSDate(aEvent.date)).getInTimezone(this._timezone);
		jsStartDateTime.timezone = this._timezone;

		var seconds;
		if (aEvent.allday) {
			jsStartDateTime.isDate = true;
			// round duration to the day
			seconds = aEvent.duration; //(Math.floor(aEvent.duration / 86400) + 1) * 86400;
		} else {
			jsStartDateTime.isDate = false;
			seconds = aEvent.duration;
		}

		calEvent.startDate = jsStartDateTime;

		if (aParentItem) {
			var jsRecIdDateTime = jsDateToDateTime(UtilsOBM.restDateToJSDate(aEvent.recurrenceId)).getInTimezone(this._timezone);
			jsRecIdDateTime.timezone = this._timezone;
			calEvent.recurrenceId = jsRecIdDateTime;
		}

		var jsEndDateTime = jsStartDateTime.clone();
		jsEndDateTime.isDate = aEvent.allday;
		var duration = Cc["@mozilla.org/calendar/duration;1"].createInstance(Ci.calIDuration);
		duration.inSeconds = seconds;
		jsEndDateTime.addDuration(duration);

		calEvent.endDate = jsEndDateTime;

		if (aEvent.category && aEvent.category != "") {
			if (calEvent.setCategories) {
				// lightning 0.9 support
				calEvent.setCategories(1, ["[OBM] " + aEvent.category]);
			} else {
				calEvent.setProperty("CATEGORIES", "[OBM] " + aEvent.category);
			}
		}

		switch (aEvent.priority) {
		case 1:
			calEvent.priority = 9;
			break;
		case 2:
			calEvent.priority = 5;
			break;
		case 3:
			calEvent.priority = 1;
			break;
		default:
			calEvent.priority = 5;
		}

		calEvent.privacy = (aEvent.privacy == 1 ? "PRIVATE": "PUBLIC");

		calEvent.removeAllAttendees();
		var attendees = aEvent.getAttendees({});
		for each(var attendeeElem in attendees) {
			this.logger.debug("_createLocalEvent: new attendee found: " + attendeeElem.email, false, false);
			var calAttendee = Cc["@mozilla.org/calendar/attendee;1"].createInstance(Ci.calIAttendee);
			calAttendee.id = "MAILTO:" + attendeeElem.email;
			calAttendee.commonName = attendeeElem.displayName;
			calAttendee.participationStatus = this._getAcceptableParticipationState(attendeeElem.status);
			switch (attendeeElem.type) {
			case "REQ":
				calAttendee.role = "REQ-PARTICIPANT";
				break;
			default:
				calAttendee.role = "OPT-PARTICIPANT";
			}
			if (attendeeElem.isOrganizer) {
				calAttendee.isOrganizer = true;
				calEvent.organizer = calAttendee;
			}
			calEvent.addAttendee(calAttendee);
		}

		var calRecurrence = this._obmRecurrenceToCalRecurrence(aEvent.recurrence, calEvent, calendar);
		if (calRecurrence) {
			calEvent.recurrenceInfo = calRecurrence;
		}

		if (calEvent.alarmOffset === undefined) {
			calEvent.clearAlarms();
		}
		if (aEvent.alert != null && aEvent.alert != -1 && aEvent.alert != 0) {
			var alert = Cc["@mozilla.org/calendar/duration;1"].createInstance(Ci.calIDuration);
			alert.inSeconds = -1 * aEvent.alert; // Alarm before event starts -> negative duration | OBM manage positive durations
			if (calEvent.alarmOffset === undefined) {
				var alarm = createAlarm();
				alarm.related = Ci.calIAlarm.ALARM_RELATED_START;
				alarm.offset = alert;
				calEvent.addAlarm(alarm);
			} else {
				calEvent.alarmOffset = alert;
			}
			if (calEvent.recurrenceInfo) {
				calEvent.alarmLastAck = now();
			}
		}

		if (aEvent.isInternal)
			calEvent.setProperty("X-OBMEXT-IS-INTERNAL", "YES");

		// FIX M#2628
		calEvent.setProperty("X-MOZ-SEND-INVITATIONS", "TRUE");

		this.logger.debug("_createLocalEvent: Event created.", false, false);
		this.logger.debug("_createLocalEvent: event has " + aEvent.getAttendees({}).length + " attendees", false, false);

		return calEvent;
	},
	_updateOrAdoptEvent: function(calendar, event, isTheFirstPass, notify) {
		this.logger.debug("_updateOrAdoptEvent()", false, false);
		this.logger.debug("  event:", false, false);
		this.logger.debug("    - title: " + event.title, false, false);
		this.logger.debug("    - uid:   " + event.uid, false, false);
		this.logger.debug("    - extId: " + event.extId, false, false);
		this.logger.debug("    - generation: " + event.generation, false, false);

		var listener = {};
		var self = this;
		//if (notify) listener.isTheFirstPass = isTheFirstPass; // Closure should be enough
		//listener.newItem = event;
		//listener.calendar = calendar;
		listener.onOperationComplete = function(aCalendar, aStatus, aOperationType, aId, aDetail) {
			//<TO_REMOVE>
			// Fix second onOperationComplete call (bug LT 0.8)
			// Fixed in LT 0.9
			try {
				if (aDetail.toString() == "Can't deduce item type based on QI") {
					dump("_updateOrAdoptEvent, Can't deduce item type based on QI\n");
					return;
				}
			} catch(e) {
				dump("_updateOrAdoptEvent, aDetail not an object?\n");
				dump(e);
			}
			//</TO_REMOVE>
			//dump("--aStatus: " + aStatus + " aOperationType: " + aOperationType + " aId: " + aId + " aDetail: " + aDetail + "\n");
			// FIXME : bug lightning ? aStatus for getItem is allways 0
			if (!Components.isSuccessCode(aStatus) || !this.localEvent) {
				if (event.status != "CANCELLED") {
					self.logger.debug("Adopt item : " + event.title, false, false);
					calendar.adoptItem(event, null);
				}
			} else {
				// Will work with LT 0.9 ... not !
				var localEvent;
				try {
					localEvent = aDetail.QueryInterface(Ci.calIEvent)
				} catch(e) {
					localEvent = this.localEvent;
				}
				if (calendar.name != localEvent.calendar.name) {
					if (event.status != "CANCELLED") {
						self.logger.debug("Adopt item : " + event.title, false, false);
						calendar.adoptItem(event, null);
					}
				} else {
					if (event.status != "CANCELLED") {
						self.logger.debug("Update item !", false, false);
						var sendInvitations = localEvent.getProperty("X-MOZ-SEND-INVITATIONS");
						if (sendInvitations == "TRUE") {
							event.setProperty("X-MOZ-SEND-INVITATIONS", "TRUE");
						}

						calendar.modifyItem(event
						/* aNewItem */
						, localEvent
						/* aOldItem */
						, null
						/* listener */
						);
					} else {
						self.logger.debug("Remove item !", false, false);
						calendar.deleteItem(localEvent, null
						/* listener */
						);
					}
				}
			}
			if (notify) self._aMergeActionCompleted(notify ? isTheFirstPass : null );
		};
		listener.onGetResult = function(aCalendar, aStatus, aItemType, aDetail, aCount, aItems) {
			if (aCount > 0) {
				this.localEvent = aItems[0];
			}
		};

		calendar.getItem(event.id, listener);
	},
	_updateParticipation: function(calendar, participation, isTheFirstPass, notify) {
		this.logger.debug("_updateParticipation:()", false, false);
		this.logger.debug("_updateParticipation: id = "+participation.id, false, false);
		this.logger.debug("_updateParticipation: extId = "+participation.extId, false, false);
		var self = this;
		this._calGetItem(calendar, participation, function (localEvent) {
			if (!localEvent) {
				// this case should not appear
				self.logger.debug("_updateParticipation: localEvent not found "+participation.id, false, false);
			} else {
				self.logger.debug("_updateParticipation: localEvent.id= " + localEvent.id, false, false);
				var newLocalEvent = localEvent.clone();
				for each (var attendee in participation.getAttendees({})) {
					// Do we really need to guard against the server sending a wrong participation status?
					self.logger.debug("_updateParticipation attendee.email:" + attendee.email, false, false);
					var participationStatus = self._getAcceptableParticipationState(attendee.status);
					var localAttendee = localEvent.getAttendeeById("MAILTO:" + attendee.email);
					if (localAttendee == null) {
						self.logger.debug("_updateParticipation: localAttendee not found", false, false);
						continue;
					}
					// Can't modifiy localAttendee because it is immutable
					var newLocalAttendee = localAttendee.clone();
					self.logger.debug("_updateParticipation actual particpationStatus:" + localAttendee.participationStatus, false, false);
					newLocalAttendee.participationStatus = participationStatus;
					self.logger.debug("_updateParticipation new status:" + newLocalAttendee.participationStatus, false, false);
					if (localAttendee.participationStatus != newLocalAttendee.participationStatus) {
						newLocalEvent.removeAttendee(localAttendee);
						newLocalEvent.addAttendee(newLocalAttendee);
						self.logger.debug("_updateParticipation: modification done.", false, false);
					}
				}
				calendar.modifyItem(newLocalEvent, localEvent, null);
			}
			if (notify) self._aMergeActionCompleted(isTheFirstPass);
		});
	},
	_getAcceptableParticipationState: function(participationState) {
		var acceptableParticipationStates = ["ACCEPTED", "DECLINED", "NEEDS-ACTION"];
		var defaultParticipationState = "NEEDS-ACTION";
		var newParticipationState = defaultParticipationState;
		for each (var state in acceptableParticipationStates) {
		if (state == participationState) {
				newParticipationState = state;
				break;
			}
		}
		return newParticipationState;
	},

	_removeEvent: function(calendar, eventId, isTheFirstPass, notify) {
		this.logger.debug("_removeEvent:()", false, false);
		this.logger.debug("  eventId: " + eventId, false, false);

		var listener = {};
		var self = this;
		listener.isTheFirstPass = isTheFirstPass; // Closure should be enough
		listener.calendar = calendar;
		listener.onOperationComplete = function(aCalendar, aStatus, aOperationType, aId, aDetail) {
			try {
				if (aDetail.toString() == "Can't deduce item type based on QI") {
					return;
				}
			} catch(e) {}

			if (aStatus == Cr.NS_OK) {
				// Found, so must be deleted
				self.logger.debug("Remove item !", false, false);
				if (this.localItem) {
					try {
						this.calendar.deleteItem(this.localItem, null
						/* listener */
						)
					} catch(e) {};
				}
			}
			if (notify) self._aMergeActionCompleted(this.isTheFirstPass);

		};
		listener.onGetResult = function(aCalendar, aStatus, aItemType, aDetail, aCount, aItems) {
			if (aCount > 0) {
				this.localItem = aItems[0];
			}
		};
		calendar.getItem(eventId, listener);
	},
	_aMergeActionCompleted: function(isTheFirstPass) {
		if (this.mergeActionCount > 0) {
			this.callerListener.onProgress(this.mergeActionIdx, this.mergeActionCount, "", "all");
			this.logger.debug("a action completed:" + this.mergeActionIdx, false, false);
		}
		this.mergeActionIdx++;
	},
	_isEventPrivateFor: function(event, calName) {
		//dump("calName: "+ calName + "\n");
		//dump("login: " + this.login + "\n");
		//dump("event.owner: "+ event.owner +"\n");
		//dump("event.privacy: "+ event.privacy +"\n");
		var attendees = event.getAttendees({});
		//dump("event.attendees: " + attendees + "\n");
		if (!event.privacy == 1) {
			return false;
		}
		if (event.owner == this.login.split("@")[0]) {
			return false;
		}
		if (!attendees) {
			return true;
		}
		for each(attendee in attendees) {
			//dump("attendee.email: " + attendee.email + "\n");
			if (attendee.email == this.login.split("@")[0]) {
				return false;
				break;
			}
		}
		return true;
	},
	mergeRemoteDiff: function(isTheFirstPass) {
		this.logger.debug("mergeRemoteDiff("+isTheFirstPass+")", false, false);
		var calMgr = Cc["@mozilla.org/calendar/manager;1"].getService(Ci.calICalendarManager);

		this.mergeActionCount = 0;
		this.mergeActionIdx = 0;
		for (var i = 0; i < this.calendars.length; i++) {
			var cal = this.calendars[i];
			var calName = calMgr.getCalendarPref_(cal, "X-OBM-NAME");
			this.mergeActionCount += this._remoteDiff[calName].updated.length;
			this.mergeActionCount += this._remoteDiff[calName].removed.length;
			this.mergeActionCount += this._remoteDiff[calName].participationChanged.length;
		}

		this.callerListener.onNewOperation("mergeRemoteDiff", this.mergeActionCount);
		this.logger.debug("Action count: " + this.mergeActionCount, false, false);

		var self = this;

		for (var i = 0; i < this.calendars.length; i++) {
			var cal = this.calendars[i];
			cal.startBatch();
			var calName = calMgr.getCalendarPref_(cal, "X-OBM-NAME");
			for each(var event in this._remoteDiff[calName].updated) {
				var calEvent = this._createLocalEvent(event, cal);
				if (this._isEventPrivateFor(event, calName) || calEvent.status == "CANCELLED") {
					this.logger.debug("event is private or refused (remove if exist)", false, false);
					this._removeEvent(cal, event.uid, isTheFirstPass, true);
				} else {
					this._updateOrAdoptEvent(cal, calEvent, isTheFirstPass, true);
				}
			}
			for each(var event in this._remoteDiff[calName].removed) {
				this._removeEvent(cal, event, isTheFirstPass, true);
			}
			for each(var participation in this._remoteDiff[calName].participationChanged) {
				this._updateParticipation(cal, participation, isTheFirstPass, true);
			}
			cal.endBatch();
			this._setLastSync(calName, this._remoteDiff[calName].lastSync);

			// we remove all local changes related to the current calendar
			// and we re-save local changes. They may be changes after the conflicts resolving
			// In case where an exception appear during the send of local change
			// we want to keep the list of changes to do after the merge of remote change
			// to replay it after.
			if (isTheFirstPass) {
				UtilsOBM.removeAllCalendarSyncItems(cal);
				for (var j in this._localDiff[cal.id].event ) {
					var ev = this._localDiff[cal.id].event[j];
					UtilsOBM.setCalendarSyncItem(cal, j, ev);
				}
			}
		}

		if (this.mergeActionCount == 0) {
			this.logger.debug("No remote event have been updated/added/removed", false, false);
			this._aMergeActionCompleted(isTheFirstPass);
		}
		this.logger.debug("Merge completed", false, false);
		this.callerListener.onOperationDone(true, null);
	},

	getLocalDiffCount : function() {
		this._localDiffOperationCount = 0;

		for ( var i in  this._localDiff ) {
			if ( this._localDiff[i].event ) {
				for ( j in this._localDiff[i].event ) {
					if ( this._localDiff[i].event[j].action ) {
						this._localDiffOperationCount++;
					}
				}
			}
		}
		//for (var i = 0; i < this.calendars.length; i++) {
			//var cal = this.calendars[i];
			//var calName = calMgr.getCalendarPref_(cal, "X-OBM-NAME");
			//this._localDiffOperationCount += this._localDiff[calName].event.removed.length;
			//this._localDiffOperationCount += this._localDiff[calName].event.updated.length;
			//this._localDiffOperationCount += this._localDiff[calName].event.added.length;
			//this._localDiffOperationCount += this._localDiff[calName].event.participationChanged.length;
		//}
		return this._localDiffOperationCount;
	},
	startBatchOnCalendars : function() {
		for (var i = 0; i < this.calendars.length; i++) {
			var cal = this.calendars[i];
			cal.startBatch();
		}
	},
	stopBatchOnCalendars : function () {
		for (var i = 0; i < this.calendars.length; i++) {
			var cal = this.calendars[i];
			cal.endBatch();
		}
	},

	_OBMCalendarsArePresent: function() {
		this.logger.debug("_ensureOBMCalendarsArePresent()", false, false);
		var calMgr = Cc["@mozilla.org/calendar/manager;1"].getService(Ci.calICalendarManager);
		var calendars = calMgr.getCalendars({});
		for each(var calendar in calendars) {
			if (calendar.type == "obm") return true;
		}
		return false;
	},

	// /////////////////////////////


	/**
	 * after retrieving the list of calendar, for each calendar, we add it into the calendar manager
	 * @TODO. NEW_OBM_CALENDAR_MANAGER. use obmICalendarsManager::subscribeObmCalendar() instead
	 */
	_subscribeTo: function(aCalInfo, readOnly) {
		this.logger.debug("_subscribeTo()", false, false);
		this.callerListener.onNewOperation("subscribingToCal", null);

		var uid = aCalInfo.uid;
		var name = UtilsOBM.getCharPref("extensions.obm.autoconf.calname", null);
		if (!name) {
			name = "[OBM] " + aCalInfo.firstname + " " + aCalInfo.lastname;
		}
		var email = aCalInfo.mail;
		var color = UtilsOBM.getRandomColor();
		var readonly = readOnly != null ? readOnly : !aCalInfo.write;
		this.logger.debug("_subscribeTo: " + uid + ", " + name + ", " + email + ", " + color + ", " + readonly, false, false);

		var calMgr = Cc["@mozilla.org/calendar/manager;1"].getService(Ci.calICalendarManager);
		UtilsOBM.setBoolPref("extensions.obm.changes.calICalendar." + uid + ".readOnly", readonly);

		var uri = Cc["@mozilla.org/network/simple-uri;1"].createInstance(Ci.nsIURI);
		uri.spec = "moz-profile-calendar://?id=" + UtilsOBM.getNextLocalCalendarID() + ";" + uid;
		var newCalendar = calMgr.createCalendar("obm", uri);
		calMgr.registerCalendar(newCalendar);
		newCalendar.name = name;
		calMgr.setCalendarPref_(newCalendar, "X-OBM-NAME", uid.toString());
		calMgr.setCalendarPref_(newCalendar, 'X-OBM-EMAIL', email.toString());
		newCalendar.setProperty("color", color.toString());
		this._setLastSync(uid, "0");
		return newCalendar;
	},
	_getCalendarsAsHash: function(calendars) {
		// Generates a hash of calendars with the OBM email as key
		var calendarsHash = {};
		var calendarManager = UtilsOBM.getCalendarManager();
		for (var i = 0; i < calendars.length ; i++) {
			var calendar = calendars[i];
			var obmEmail = calendarManager.getCalendarPref_(calendar, "X-OBM-EMAIL");
			calendarsHash[obmEmail] = calendar;
		}
		return calendarsHash;
	},

	mergeOneCalendar: function(aCalendar, aListener) {
		this.logger.debug("mergeOneCalendar()", false, false);
		this.logger.debug("  aCalendar: " + aCalendar, false, false);
		var sch = this._merge(aListener, aCalendar);
		sch.taskName="mergeOneCalendar scheduler";
		return sch;
	},

	mergeAllCalendars: function(aListener) {
		this.logger.debug("mergeAllCalendars()", false, false);
		var sch = this._merge(aListener, false);
		sch.taskName="mergeAllCalendars scheduler";
		return sch;
	},

	_merge: function(aListener, singleCalendar) {
		
		this.logger.debug("mergeCalendars", false, false);
		this.callerListener = aListener;
		this.callerListener.onStart();
		if (UtilsOBM.getBoolPref("extensions.obm.calEvents.is_sync_enabled", true) == false) {
			this.callerListener.onStop(Cr.NS_OK, "done - Events not sync-able");
			return null;
		}

		this.scheduler = new Scheduler();
		this.scheduler.schedule(new loginTask(this, true), true);
		if (this._OBMCalendarsArePresent()) {
		  this.scheduler.schedule(new calendarMetaDataTask(this), true);
		}
		else {
		  this.scheduler.schedule(new listCalendarTask(this, true), true);
		}
		this.scheduler.schedule(new listCategoriesTask(this), true);
		this.scheduler.schedule(new calendarsRemoteChangesTask(this, singleCalendar, true), true);
		this.scheduler.schedule(new calSendLocalDiffTask(this), true);

		// we're waiting more than one second because mysql, on the server side,
		// cannot provide a diff date lower than a second 
		this.scheduler.schedule(new sleepTask(1100, aListener), true);
		this.scheduler.schedule(new calendarsRemoteChangesTask(this, singleCalendar, false), true);

		this.scheduler.scheduleOnEnd(function(){
			aListener.onStop(Cr.NS_OK, "done");
		});
		this.scheduler.scheduleOnError(function(error){
			aListener.onStop(Cr.NS_ERROR_FAILURE, error);
		});
		return this.scheduler;
	},

	getSubscribleCalendars: function(aListener) {
		this.logger.debug("getSubscribleCalendars()", false, false);
		this.callerListener = aListener;
		this.callerListener.onStart();

		if (UtilsOBM.getBoolPref("extensions.obm.calEvents.is_sync_enabled", true) == false) {
			this.callerListener.onStop(Cr.NS_OK, "done - Events not sync-able");
			return null;
		}

		this.scheduler = new Scheduler();
		this.scheduler.schedule(new loginTask(this), true);
		this.scheduler.schedule(new listCalendarTask(this, false), true);
		this.scheduler.scheduleOnEnd(function () {
			aListener.onStop(Cr.NS_OK, "done");
		});
		return this.scheduler;
	},

	purge: function(calName, aListener) {
		this.logger.debug("purge(" + calName + ")", false, false);

		this._calNameToPurge = calName;
		this.callerListener = aListener;
		this.callerListener.onStart();

		this.scheduler = new Scheduler();
		this.scheduler.schedule(new loginTask(this, true), true);
		this.scheduler.schedule(new purgeCalendarTask(this, calName), true);
		this.scheduler.scheduleOnEnd(function () {
			aListener.onStop(Cr.NS_OK, "done");
		});
		return this.scheduler;
	}
};

