/* ***** BEGIN LICENSE BLOCK *****
 *
 * Copyright (c) 2011 Groupe LINAGORA
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of the
 *  License, (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  http://www.obm.org/
 *
 * ***** END LICENSE BLOCK ***** */


/*
 This file defines all asynchronous task to perform synchronization of
 calendar data between lightning and obm-sync.
 These tasks should be used with the Scheduler object.
*/

function listCalendarTask(calSync, subscribeOwnerCalendar) {
	this.taskName = "listCalendarTask";
	this.wsMethod = 'calendar/listCalendars';
    this.subscribeOwnerCalendar = subscribeOwnerCalendar;
	this.calSync = calSync;
}

listCalendarTask.prototype = new xhrTask();

listCalendarTask.prototype.preRequest = function(data, parameters) {
	this.calSync.callerListener.onNewOperation("listCalendars", null);
	parameters.add("sid",  this.calSync.session_id);
};

listCalendarTask.prototype.postRequest = function(callback, data, xml, error) {
	if (error) {
        callback(this._processError(error));
		return;
	}
	else {
        var calendars = new CalendarArrayShell(xml);

        if (this.subscribeOwnerCalendar) {
            var count = calendars.length;
            var login = {};
            UtilsOBM.getOBMCredentials(login, {});
            login = login.value.split('@')[0];
            for (var i = 0; i < count; i++) {
                var calInfo = calendars.getElementAt(i);
                if (calInfo.uid == login) {
                    //TODO. NEW_OBM_CALENDAR_MANAGER. use obmICalendarsManager::subscribeObmCalendar() instead
                    // We consider that the user can always read/write on its own calendar
					var readOnly = false;
					this.calSync._subscribeTo(calInfo, readOnly);
//                     this.calSync._subscribeTo(calInfo);
                    break;
                }
            }
            this.calSync.callerListener.onOperationDone(true, null);
        }
        else 
            this.calSync.callerListener.onOperationDone(true, calendars);
	}
	callback();
};



/**
 * task that retrieves calendar informations
 */
function calendarMetaDataTask(calSync) {
	this.taskName = "calendarMetaDataTask";
	this.wsMethod = 'calendar/getCalendarMetadata';
	this.calSync = calSync;
}

calendarMetaDataTask.prototype = new xhrTask();

calendarMetaDataTask.prototype.preRequest = function(data, p) {
	this.calSync.callerListener.onNewOperation("getCalendarMetadata", {});

	p.add("sid",  this.calSync.session_id);

    var calendarEmails = this.calSync._getCalendarEmails();
    for (var i = 0 ; i < calendarEmails.length ; i++)
        p.add("calendar", calendarEmails[i]);

};

calendarMetaDataTask.prototype.postRequest = function(callback, data, xml, error) {
    this.calSync.callerListener.onOperationDone(true, null);
	if (error) {
		// we do nothing. let's ignore the error
	}
	else {
        var calendars = new CalendarArrayShell(xml);
        this.calSync.processCalendarMetadata(calendars);
	}
	callback();
};


/**
 * task that retrieves list of categories
 */
function listCategoriesTask(calSync) {
	this.taskName = "listCategoriesTask";
	this.wsMethod = 'calendar/listCategories';
	this.calSync = calSync;
}

listCategoriesTask.prototype = new xhrTask();

listCategoriesTask.prototype.preRequest = function(data, parameters) {
	this.calSync.callerListener.onNewOperation("listCategories", {});
	parameters.add("sid",  this.calSync.session_id);
};

listCategoriesTask.prototype.postRequest = function(callback, data, xml, error) {
	if (error) {
		callback(this._processError(error));
		return;
	}
	else {
        this.calSync.callerListener.onOperationDone(true, null);
        var result = new GetCategoriesShell(xml);
        var serverCategories = result.getCategories();

		var currentCategories = getLocalizedPref("calendar.categories.names");
		if (currentCategories && currentCategories != "") {
			currentCategories = currentCategories.split(",");
		} else {
			currentCategories = [];
		}

		var newCategories = [];
		for each (var category in currentCategories) {
			if ( !category.match(/\[OBM\]/)) {
				newCategories.push(category);
			}
		}

        for(var i=0; i< serverCategories.length; i++) {
			var label = "[OBM] " + serverCategories[i].label;
			newCategories.push(label);
        }
		setLocalizedPref("calendar.categories.names", newCategories.join(","));
	}
	callback();
};



/**
 * task lauching other tasks to retrieve changes of all given calendars
 */
function calendarsRemoteChangesTask(calSync, singleCalendar, isFirstPass) {
	this.taskName = "calendarRemoteChangesTask";
    this.singleCalendar = singleCalendar;
    this.isFirstPass = isFirstPass;
	this.calSync = calSync;
}

calendarsRemoteChangesTask.prototype = new basicTask();

calendarsRemoteChangesTask.prototype.run = function(callback, data) {

    var calendars = [];
    if (this.singleCalendar)
        calendars.push(this.singleCalendar);
    else
        calendars = this.calSync.getSubscribedOBMCalendar();
    this.calSync.loadLocalDiff(calendars);

	this.calSync.callerListener.onNewOperation("getSyncWithSortedChanges", null);


    var calMgr = Cc["@mozilla.org/calendar/manager;1"].getService(Ci.calICalendarManager);
    var tasks = [];
	for (var i = 0; i < calendars.length; i++) {
        var cal = calendars[i];
        var calName = calMgr.getCalendarPref_(cal, "X-OBM-NAME");
        tasks.push(new singleCalendarRemoteChangesTask(this.calSync, cal, calName, this.isFirstPass, i, calendars.length));
    }

    var self = this;
    var calRemoteChangesEndTask = {
		taskName : "calRemoteChangesEndTask",
        run: function(callback, data) {
            self.calSync.callerListener.onOperationDone(true, null);
            self.calSync.mergeRemoteDiff(self.isFirstPass);
            callback();
        },
        abort: function() {}
    }
    tasks.push(calRemoteChangesEndTask);

	callback(null, {tasks: tasks});
};




function singleCalendarRemoteChangesTask(calSync, calendar, calName, isFirstPass, idx, total) {
	this.taskName = "singleCalendarRemoteChangesTask";
	this.wsMethod = 'calendar/getSyncWithSortedChanges';
    this.calendar = calendar;
    this.calName = calName;
    this.isFirstPass = isFirstPass;
    this.idx = idx;
    this.total = total;
	this.calSync = calSync;
}

singleCalendarRemoteChangesTask.prototype = new xhrTask();

singleCalendarRemoteChangesTask.prototype.preRequest = function(data, parameters) {
    this.calSync.callerListener.onProgress(this.idx, this.total, "", this.calName);
	parameters.add("sid",  this.calSync.session_id);
    parameters.add("calendar", this.calName);
	parameters.add("lastSync", this.calSync._getLastSync(this.calendar));
};

singleCalendarRemoteChangesTask.prototype.postRequest = function(callback, data, xml, error) {
	if (error) {
		callback(this._processError(error));
		return;
	}
	else {
        var changes = new GetSyncEventWithSortedChangesResponseShell(xml);
		this.logger.debug("changes lastSync : "+changes.lastsync, false, false);
        this.calSync.storeRemoteDiff(this.calendar, this.calName, changes);

        if (this.isFirstPass)
            this.calSync.resolveConflicts(this.calName);
	}
	callback();
};


/**
 *
 */
function calSendLocalDiffTask(calSync) {
	this.taskName = "calSendLocalDiffTask";
	this.calSync = calSync;
}

calSendLocalDiffTask.prototype = new basicTask();

calSendLocalDiffTask.prototype.run = function(callback, data) {

    var localDiffCount = this.calSync.getLocalDiffCount();
    this.calSync.startBatchOnCalendars();

    this.calSync.callerListener.onNewOperation("sendLocalDiff", localDiffCount);

    var tasks = [];
    var calMgr = Cc["@mozilla.org/calendar/manager;1"].getService(Ci.calICalendarManager);

    //this.logger.debug("this._localDiff contents: "+JSON.stringify(this.calSync._localDiff),false,false);

    var idx = 0;
    for (var i = 0; i < this.calSync.calendars.length; i++) {
        var cal = this.calSync.calendars[i];
        var calName = calMgr.getCalendarPref_(cal, "X-OBM-NAME");

        for ( var j in this.calSync._localDiff[cal.id].event ) {
            var evChange = this.calSync._localDiff[cal.id].event[j];

            if ( evChange.action == "removed" ) {

                var t = new removedEventTask(this.calSync, cal, calName, j, this.calSync._supportExtId, idx, localDiffCount);
                t.notification = evChange.notification ? evChange.notification : false;
                t.generation = evChange.generation ? evChange.generation : 0;

                tasks.push(t);

            } else if (evChange.action == "added") {
                var cet = new createEventTask(this.calSync, cal, calName, j, idx, localDiffCount);
                cet.notification = evChange.notification ? evChange.notification : false;

                var t = new getCalendarItemTask(cal, j);
                tasks.push(t);
                tasks.push(cet);

            } else if (evChange.action == "updated" ) {
                var met = new modifyEventTask(this.calSync, cal, calName, j, idx, localDiffCount);
                met.notification = evChange.notification ? evChange.notification : false;

                var t = new getCalendarItemTask(cal, j);
                tasks.push(t);
                tasks.push(met);

            } else if ( evChange.action == "participationChanged" ) {
                var pet = new participationChangedTask (this.calSync, cal, calName, j, idx, localDiffCount);
                pet.notification = evChange.notification ? evChange.notification : false;
                pet.generation = evChange.sequence ? evChange.sequence : 0;
                var t = new getCalendarItemTask(cal, j);
                tasks.push(t);
                tasks.push(pet);
            }
            idx++;
        }
    }

    var self = this;
    var calSendLocalDiffEndTask = {
		taskName:"calSendLocalDiffEndTask",
        run: function(callback, data) {
            self.calSync.callerListener.onOperationDone(true, null);
            self.calSync.stopBatchOnCalendars();
            callback();
        },
        abort: function() {}
    }
    tasks.push(calSendLocalDiffEndTask);

    callback(null, {tasks: tasks});
}




function removedEventTask(calSync, calendar, calName, id, isExtId, idx, total) {
	this.taskName = "removedEventTask";
    if (isExtId)
        this.wsMethod = 'calendar/removeEventByExtId';
    else
        this.wsMethod = 'calendar/removeEvent';

    this.calendar = calendar;
    this.calName = calName;
    this.id = id;
    this.isExtId = isExtId;
    this.idx = idx;
    this.total = total;
    this.notification = false;
    this.generation = 0;
	this.calSync = calSync;
}

removedEventTask.prototype = new xhrTask();

removedEventTask.prototype.preRequest = function(data, parameters) {
    this.calSync.callerListener.onProgress(this.idx, this.total, "removed", this.calName);
	parameters.add("sid",  this.calSync.session_id);
    parameters.add("calendar", this.calName);
    if (this.isExtId)
        parameters.add("extId", this.id);
    else
        parameters.add("id", this.id);
	parameters.add("notification", this.notification);
	parameters.add("sequence", this.generation);
};

removedEventTask.prototype.postRequest = function(callback, data, xml, error) {
	if (error) {
		this.calSync.callerListener.onError(this._processError(error));
	}
	else {
        UtilsOBM.removeCalendarSyncItem(this.calendar, this.id);
	}
	callback();
};



function getCalendarItemTask(calendar, id) {
    this.calendar = calendar;
    this.id = id;
	this.taskName = "getCalendarItemTask";
}

getCalendarItemTask.prototype = new basicTask();

getCalendarItemTask.prototype.run = function(callback, data) {
    var self = this;
    var listener = {
        onGetResult : function(aCalendar, aStatus, aItemType, aDetail, aCount, aItems) {
			if (aCount > 0) {
				this.localEvent = aItems[0];
			}
		},
        onOperationComplete : function(aCalendar, aStatus, aOperationType, aId, aDetail) {
            if (Components.isSuccessCode(aStatus)) {
                callback(null, {data: this.localEvent} );
            }
            else {
                callback("Error during retrieving an item");
            }
        }
    };
    this.calendar.getItem(this.id, listener);
};

getCalendarItemTask.prototype.abort = function() {

}


function createEventTask(calSync, calendar, calName, id, idx, total) {
	this.taskName = "createEventTask";
    this.wsMethod = 'calendar/createEvent';

    this.calendar = calendar;
    this.calName = calName;
    this.id = id;
    this.idx = idx;
    this.total = total;
    this.notification = false;
    this.generation = 0;
	this.calSync = calSync;
	this.event = null; // eventually fill it if it is not given by the previous task
}

createEventTask.prototype = new xhrTask();

createEventTask.prototype.preRequest = function(event, parameters) {
	this.event = this.event || event;
    this.calSync.callerListener.onProgress(this.idx, this.total, "added", this.calName);

    this.generation = this.event.generation;

    var ev = this.calSync._mozEventToOBMEvent(this.calendar, this.calName, this.event, false);
    
	parameters.add("sid",  this.calSync.session_id);
    parameters.add("calendar", this.calName);
	parameters.add("event", JSObjEventShell(ev));
	parameters.add("notification", this.notification);
};

createEventTask.prototype.postRequest = function(callback, event, xml, error) {
	if (error) {
        this.calSync.stopBatchOnCalendars();
        callback(this._processError(error));
        return;
	}
	else {

        var newId = _xS(xml.*::value);
        var newLocalEvent = this.event.clone();
		UtilsOBM.setOBMEventId(this.calSync._supportExtId, newLocalEvent, newId);
		this.calendar.modifyItem(newLocalEvent, this.event, null);

        UtilsOBM.removeCalendarSyncItem(this.calendar, this.id);
    	callback();
	}
};



function modifyEventTask(calSync, calendar, calName, id, idx, total) {
	this.taskName = "modifyEventTask";
    this.wsMethod = 'calendar/modifyEvent';
    this.calendar = calendar;
    this.calName = calName;
    this.id = id;
    this.idx = idx;
    this.total = total;
    this.notification = false;
    this.generation = 0;
    this.event = null; // eventually fill it if it is not given by the previous task
	this.calSync = calSync;
}

modifyEventTask.prototype = new xhrTask();

modifyEventTask.prototype.preRequest = function(event, parameters) {
	this.event = this.event || event;
    this.calSync.callerListener.onProgress(this.idx, this.total, "updated", this.calName);

    this.generation = this.event.generation;

    var ev = this.calSync._mozEventToOBMEvent(this.calendar, this.calName, this.event, false);

	parameters.add("sid",  this.calSync.session_id);
    parameters.add("calendar", this.calName);
    parameters.add("event", JSObjEventShell(ev));
    parameters.add("updateAttendees", true);
	parameters.add("notification", this.notification);
	parameters.add("sequence", this.generation);

};

modifyEventTask.prototype.postRequest = function(callback, event, xml, error) {
	if (error) {
        this.calSync.stopBatchOnCalendars();
        callback(this._processError(error));
        return;
	}
	else {
        /*
        var event = new EventShell(xml);
		var calEvent = this.calSync._createLocalEvent(event, this.calendar);
		this.calSync._updateOrAdoptEvent(this.calendar, calEvent, null, false);
        */
        UtilsOBM.removeCalendarSyncItem(this.calendar, this.id);
        callback();
	}
};


function participationChangedTask (calSync, calendar, calName, id, idx, total) {
	this.taskName = "participationChangedTask";
	this.wsMethod = 'calendar/changeParticipationState';
    this.calendar = calendar;
    this.calName = calName;
    this.id = id;
    this.idx = idx;
    this.total = total;
    this.notification = false;
    this.generation = 0;
    this.event = null; // eventually fill it if it is not given by the previous task
	this.calSync = calSync;
}

participationChangedTask.prototype = new xhrTask();

participationChangedTask.prototype.preRequest = function(event, parameters) {
	this.event = this.event || event;
    if (!this.event)
        return true;

    this.calSync.callerListener.onProgress(this.idx, this.total, "participationChanged", this.calName);

    var calMgr = Cc["@mozilla.org/calendar/manager;1"].getService(Ci.calICalendarManager);
	var localAttendeeEmail = calMgr.getCalendarPref_(this.calendar, "X-OBM-EMAIL");
	var localAttendeeId = "MAILTO:" + localAttendeeEmail;
    var localAttendee = null;
    for each (var att in this.event.getAttendees({})) {
        //this.logger.debug("Parsing attendee: " + att.id, false, false);
        if (att.id.toLowerCase() == localAttendeeId.toLowerCase()) {
            localAttendee = att;
        }
    }
    if (localAttendee == null) {
        this.logger.error("participationChangedTask: Could not find attendee '" + localAttendeeId + "' in order to tell server that his participation status was updated!", false, false);
        return true;
    }

    this.participationStatus = localAttendee.participationStatus;

	parameters.add("sid",  this.calSync.session_id);
    parameters.add("calendar", this.calName);
    parameters.add("extId", this.id);
	parameters.add("state", localAttendee.participationStatus);
	parameters.add("notification", this.notification);
	parameters.add("sequence", this.generation);
	return false;
};

participationChangedTask.prototype.postRequest = function(callback, data, xml, error) {
	if (error) {
		this.calSync.callerListener.onError(this._processError(error));
	}
	else {
        UtilsOBM.removeCalendarSyncItem(this.calendar, this.id);
	}
	callback();
};





function purgeCalendarTask(calSync, calName) {
	this.taskName = "purgeCalendarTask";
	this.wsMethod = 'calendar/purge';
	this.calName = calName;
	this.calSync = calSync;
}

purgeCalendarTask.prototype = new xhrTask();

purgeCalendarTask.prototype.preRequest = function(data, parameters) {
	parameters.add("sid",  this.calSync.session_id);
	parameters.add("calendar", this.calName);
};

purgeCalendarTask.prototype.postRequest = function(callback, data, xml, error) {
	if (error) {
        callback(this._processError(error));
		return;
	}
	else {
        this.logger.notice("Purge: obm-sync performed purged. Resetting connector", false, false);
		callback();
	}
};






// --------------- utilities objects to format data request

function _eventShell(aEvent, aKind) {
  aEvent.QueryInterface(Ci.obmIREventRecurrence);
  if (aKind == "event") {
		var obj = <event/>; // xmlns="http://www.obm.org/xsd/sync/event.xsd"/>;
		if (aEvent.uid) {
			obj.@id = aEvent.uid;
		}
		if (aEvent.isInternal)
			obj.@isInternal = "true";
		else
			obj.@isInternal = "false";

		if ( "generation" in aEvent && aEvent.generation !== null ) {
			obj.@sequence = aEvent.generation ? parseInt(aEvent.generation,10) : 0 ;
		} else {
			obj.@sequence = 0;
		}

		_appendXMLElement(obj, "extId", aEvent.extId);
  } else {
		var obj = <eventException/>
		_appendXMLElement(obj, "recurrenceId", aEvent.recurrenceId);
  }

  obj.@allDay = aEvent.allday;
  obj.@type = "VEVENT";

  _appendXMLElement(obj, "owner", aEvent.owner);
  _appendXMLElement(obj, "priority" , aEvent.priority);
  _appendXMLElement(obj, "privacy" , aEvent.privacy);
  _appendXMLElement(obj, "date" , aEvent.date);
  _appendXMLElement(obj, "duration" , aEvent.duration);
  _appendXMLElement(obj, "title" , aEvent.title);
  _appendXMLElement(obj, "description" , aEvent.description);
  _appendXMLElement(obj, "category" , aEvent.category);
  _appendXMLElement(obj, "location" , aEvent.location);
  _appendXMLElement(obj, "alert" , aEvent.alert);
  _appendXMLElement(obj, "opacity" , aEvent.opacity);
  _appendXMLElement(obj, "ownerEmail", aEvent.ownerEmail);

  var atts = aEvent.getAttendees({});
  for (var i = 0; i < atts.length; i++) {
	var att = <attendee email={atts[i].email} displayName={atts[i].displayName} state={atts[i].status} required={atts[i].type}/>;
	if (atts[i].isOrganizer)
		att.@isOrganizer ='true';
    obj.attendees.attendee[i] = att;
  }

  obj.recurrence = <recurrence kind={aEvent.kind} freq={aEvent.frequence} days={aEvent.days}/>;
  if (aEvent.end)
  	obj.*::recurrence.@end = aEvent.end;

  var exs =	aEvent.getExceptions({});
  for (var i = 0; i < exs.length; i++) {
  	obj.recurrence.exceptions.exception[i] = exs[i];
  }

  var eventExs = aEvent.getEventExceptions({});
  for (var i = 0; i < eventExs.length; i++) {
  	obj.recurrence.eventExceptions.eventException[i] = _eventShell(eventExs[i], "eventException");
  }

  return obj;
}

/**
 * @param obmIREvent aEvent
 */
function JSObjEventShell(aEvent) {
  var obj = _eventShell(aEvent, "event");
  return ("<?xml version=\"1.0\" encoding=\"UTF-8\"?>" + obj.toString());
}