/* ***** BEGIN LICENSE BLOCK *****
 *
 * Copyright (c) 2011 Groupe LINAGORA
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of the
 *  License, (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  http://www.obm.org/
 *
 * ***** END LICENSE BLOCK ***** */


// utilities functions

function bind(f, obj) {
  return function() {
    return f.apply(obj, arguments);
  }
}

function process(fct,thisp) {
   this.run = thisp ?  function() { fct.apply(thisp,arguments) } : fct;
}

function callAsync(fct,thisp) {
    var threadManager = Components.classes["@mozilla.org/thread-manager;1"]
        .getService(Components.interfaces.nsIThreadManager);
    threadManager.mainThread.dispatch(new process(fct,thisp),0);
}


/**
 * the chief orchestra allowing to run asynchronous task one after an other
 */
function Scheduler(){
    this.tasks = [];    
    this.executing = false;
    this.error = null;
    this.callbackOnEnd = [];
    this.callbackOnError = [];
    this._stopped = false;
    this.taskName = "Scheduler";
}
Scheduler.prototype = {

	get logger() {
		if (!this._logger) {
			this._logger = Cc["@obm.org/calendar/logger;1"].createInstance(Ci.obmILogger);
			var writer = Cc["@obm.org/calendar/writer;1?type=unixfile"].getService(Ci.obmILoggerWriter);
			this._logger.setLevelWriter(this.logger.LOAD_LEVEL, writer);
			this._logger.setHeader("Scheduler: ");
		}
		return this._logger;
	},

    /**
     * add a task to the list of task to execute
     * @param basicTask task the task to add
     * @param boolean append  true to append the task, or false to put it at the beginning
     */
    schedule : function (task, append) {
        this._stopped = false;
        if (append)
            this.tasks.push(task);
        else
            this.tasks.unshift(task);
    },

    /**
     * add a list of tasks to the list of task to execute
     * @param array tasks the list of basicTask to add
     * @param boolean append  true to append the tasks, or false to put it at the beginning
     */
    scheduleList : function (tasks, append) {
      this._stopped = false;
      if (append) {
        this.tasks = this.tasks.concat(tasks);
      }
      else {
        this.tasks = tasks.concat(this.tasks);
      }
    },

    /**
     * declares a callback to call when all tasks are finished
     * or when the scheduler is stopped without errors
     * @param Function fct  the callback to call. should take a Scheduler as first argument
     * @param boolean append  true to append the callback, or false to put it at the beginning of the list of callback
     */
    scheduleOnEnd : function (fct, append) {
      if (append)
        this.callbackOnEnd.push(fct);
      else
        this.callbackOnEnd.unshift(fct);
    },

    /**
     * declares a callback to call when an error occured in a task 
     * or when the scheduler is stopped with an error
     * @param Function fct  the callback to call. should take a Scheduler as first argument
     * @param boolean append  true to append the callback, or false to put it at the beginning of the list of callback
     */
    scheduleOnError : function (fct, append) {
      if (append)
        this.callbackOnError.push(fct);
      else
        this.callbackOnError.unshift(fct);
    },

    /**
     * execute the first task in the list. The task is then removed from the list
     * @param mixed param parameter to pass to the run method of the next task
     *                   this parameter can be passed by the previous task
     *                   by setting a "data" member in the response object given
     *                   to the callback.
     */
    next : function(param) {
        //this.logger.debug("next", false, false);
        //if (this._executing) return; // avoid reentrant call
        //this.executing = true;
        //this.logger.debug("executing", false, false);

        if (this.tasks.length) {
            var t = this.tasks.shift();
            var self = this;
            this.logger.debug("call async. task="+t.taskName, false, false);
            callAsync(function() {
                try {
                  t.run(function(error, response){
                    // the callback which process the response of the task
                    if (error) {
                      self.stop(error);
                    }
                    else if(response === false) {
                      self.stop();
                    }
                    else {
                      if (response && response.tasks) {
                        self.scheduleList(response.tasks, response.appendTasks);
                      }
                      if (response && response.data) {
                        // the given data will be parameters of the next task
                        self.next(response.data);
                      } else {
                        self.next();
                      }
                    }
                  }, param);
                }
                catch(e) {
                    //this.logger.debug("error in task: "+e, false, false);
                    // we have an error during the execution of the task
                    t.abort();
                    self.stop(e);
                }
            }, this);
        }
        else {
          //this.logger.debug("no more tasks", false, false);
          this.stop();
        }

        //this.executing = false;
    },

    /**
     * stop the execution of the scheduler. it removes all task
     * from the task list, and call callbacks "onEnd" or "onError"
     */
    stop : function(error) {
        if (this._stopped) {
          //this.logger.debug("stop (already stopped)", false, false);
          return;
        }
        if (error)
          this.error = error;
        //this.logger.debug("stop", false, false);
        this.tasks = [];
        this._stopped = true;
        if (this.error) {
          for(var i=0; i < this.callbackOnError.length;i++)
            this.callbackOnError[i](this.error);
        }
        else {
          for(var i=0; i < this.callbackOnEnd.length;i++)
            this.callbackOnEnd[i]();
        }
    },

    // ----- task interface. it allow to add a scheduler as a task in an other
    // scheduler.

    run: function (callback, data) {
      //this.logger.debug("run sub scheduler", false, false);
      var self = this;
      
      this.scheduleOnEnd(function(){
        //self.logger.debug("end of sub scheduler", false, false);
        callback();
      });

      this.scheduleOnError(function(error){
        //self.logger.debug("end of sub scheduler (error)", false, false);
        callback(error);
      });

      this.next(data);
    },
    abort : function () {
      this.stop();
    }
}
