/* ***** BEGIN LICENSE BLOCK *****
 *
 * Copyright (c) 2011 Groupe LINAGORA
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of the
 *  License, (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  http://www.obm.org/
 *
 * ***** END LICENSE BLOCK ***** */

/**
 * base object for tasks. Tasks should be inherit from it
 */
function basicTask() {
	this.taskName = "basicTask";
}

basicTask.prototype = {

	get stringBundle () {
		if (!this._sb) {
			this._sb = Cc["@mozilla.org/intl/stringbundle;1"]
				.getService(Ci.nsIStringBundleService)
				.createBundle("chrome://obm-extension/locale/calendar/errors_warnings.properties");
		}
		return this._sb;
	},

	get logger() {
		if (!this._logger) {
			this._logger = Cc["@obm.org/calendar/logger;1"].createInstance(Ci.obmILogger);
			var writer = Cc["@obm.org/calendar/writer;1?type=unixfile"].getService(Ci.obmILoggerWriter);
			this._logger.setLevelWriter(this.logger.LOAD_LEVEL, writer);
			this._logger.setHeader("task "+this.taskName+": ");
		}
		return this._logger;
	},

	/**
	 * It executes the task itself. you should redefine it in descendant object
	 * to implement what you want to do. The code executed by this method can
	 * be synchrone or asynchrone. In both case, it must call the callback function
	 * when the task is finished, so the next task in the stack will be called.
	 * You can indicate an object containing options for the scheduler, as a second
	 * parameter (the first argument should be null then).
	 * In case of errors, it can call the callback with the error as first argument, to indicate
	 * that next tasks won't be executed.
	 * @param function the callback to call when the task is finished
	 * @param mixed data  the data hold by the scheduler. can be an object (calendarSync for example)
	 */
	run : function(callback, data) {
		
	},
	/**
	 * called by the schedule when an exception appeared during the
	 * execution of the task.
	 */
	abort : function() {
	},

	/**
	 * utility method to retrieve the message corresponding to the given error
	 */
	_processError : function(error) {
		var printError;
		this.logger.debug("_processError: "+error);
		var msg = Cc["@mozilla.org/supports-string;1"].createInstance(Ci.nsISupportsString);
		msg.data = error.message;
		switch(error.fileName) {
			case "CONNECTION_FAILURE":
				printError = "connectFail";
				break;
			case "OBM_ERROR":
				if (error.message == "Invalid access token")
					printError = "invalidToken";
				else
					printError = "srvError";
				break;
			case "BAD_RESPONSE_VALUE":
			default:
				printError = "srvError";
				break;
		}

		try {
			printError = this.stringBundle.GetStringFromName("obm.syncerror." + printError);
		} catch(e) {
			//printError = printError;
		}
		return printError;
	}
}


/**
 * base class for objects that use asynchronous call to xmlhttprequest
 */
function xhrTask() {
	this.xhr = null;
	this.taskName = "xhrTask";
	this.wsMethod = '';
}

xhrTask.prototype = new basicTask();


xhrTask.prototype.abort = function() {
    if (this.xhr) {
        this.xhr.abort();
        this.xhr = null;
    }
}

xhrTask.prototype.run = function (callback, data) {
	this.logger.debug("run start", false, false);
	var RESTClient = Cc["@obm.org/restclient;1"].createInstance().wrappedJSObject;
	var p = RESTClient.newRESTClientParameters();

	if(this.preRequest(data, p)) {
		callback();
		return;
	}

	var self = this;
	this.xhr = RESTClient.send(this.wsMethod, p,
		function(xml, error) {
			try {
				self.logger.debug(self.wsMethod+" response: "+xml,false,false);
				self.postRequest(callback, data, xml, error);
			}
			catch(e) {
				callback(e);
			}
		});
}

xhrTask.prototype.preRequest = function(data, parameters) {

}

xhrTask.prototype.postRequest = function(callback, data, aResult, aError) {
	callback();
}

/**
 * a task which is waiting for a given number of millisecond
 */
function sleepTask(sleeptime, listener) {
    this.timer = null;
    this.sleeptime = sleeptime;
	this.taskName = "sleepTask";
	this.listener = listener;
}

sleepTask.prototype = new basicTask();

sleepTask.prototype.run = function(callback, params) {

    this.timer = Cc["@mozilla.org/timer;1"].createInstance(Ci.nsITimer);

    this.logger.debug("timer init", false, false);
    this.listener.onNewOperation("wait-one-seconde", null);
    var self = this;
    this.timer.initWithCallback({
            notify: function(timer) {
                self.logger.debug("ok, time is out", false, false);
                self.listener.onOperationDone(true, null);
                self.timer = null;
                callback();
            }
        },
        this.sleeptime,
        this.timer.TYPE_ONE_SHOT);
};

sleepTask.prototype.abort = function() {
    if (this.timer) {
        this.timer.cancel();
        this.timer = null;
    }
}


function loginTask(syncObj, checkSyncVersion) {
	this.taskName = "loginTask";
	this.checkSyncVersion = checkSyncVersion;
	this.syncObj = syncObj;
}

loginTask.prototype = new xhrTask();

loginTask.prototype.run = function(callback, params) {
	this.logger.debug("run start", false, false);
	if (this.syncObj)
		this.syncObj.callerListener.onNewOperation("logUserIn", null);

	var aLogin = {};
	var aPassword = {};
	if (!UtilsOBM.getOBMCredentials(aLogin, aPassword)
		|| !aLogin.value || aLogin.value == ""
		|| !aPassword.value || aPassword.value == "") {
		if (this.syncObj)
			this.syncObj.callerListener.onError("No login info");
		throw ("No login info");
	}
	this.syncObj.login = aLogin.value;

	var RESTClient = Cc["@obm.org/restclient;1"].createInstance().wrappedJSObject;
	var p = RESTClient.newRESTClientParameters();
	p.add("login", 		aLogin.value);
	p.add("password", 	aPassword.value);
	p.add("origin", 	UtilsOBM.getOrigin());

	var self = this;
	this.xhr = RESTClient.send("login/doLogin", p,
		function(xml, error) {
			if (self.syncObj)
				self.syncObj.callerListener.onOperationDone(true, null);
			else
				self.logger.debug("WARNING!!! no syncObj", false, false);

			if (error) {
				var printError;
				var msg = Cc["@mozilla.org/supports-string;1"].createInstance(Ci.nsISupportsString);
				if (error.fileName == "OBM_ERROR") {
					msg.data = "No token available";
					printError = "authFault";
				} else {
					msg.data = error.message;
					printError = "connectFail";
				}
				var username = {};
				var password = {};
				if (UtilsOBM.getOBMCredentials(username, password)) {
					UtilsOBM.setOBMCredentials(username.value, null);
					UtilsOBM.setTemporaryPassword(null);
				}
				try {
					printError = self.stringBundle.GetStringFromName("obm.syncerror." + printError);
				} catch(e) {
					printError = printError;
				}
				callback(printError);
				return;
			}
			var sid = _xS(xml.*::sid);
			if (self.syncObj)
				self.syncObj.session_id = sid;
			self.logger.debug("token obtained with sid:"+sid, false, false);

			if (xml.*::version != undefined) {
				if (self.syncObj) {
					self.syncObj.syncVersion = xml.*::version.@major
									+ "." + xml.*::version.@minor
									+ "." + xml.*::version.@release;
					self.logger.debug("sync version:" + self.syncObj.syncVersion, false, false);
				}

				if (self.checkSyncVersion && self.syncObj && !self.syncObj.initServerCapabilitiesFromSyncVersion() ) {
					callback("RE_INIT_NEEDED");
					return;
				}
			}

			var domain = _xS(xml.*::domain);
			UtilsOBM.setCharPref("extensions.obm.server.domain", domain);

			callback();
		});

};

